"""
Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
Copyright (c) 2011-2013 NYU                      (Clement Farabet)
Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. Neither the names of Xilinx, Facebook, Deepmind Technologies, NYU,
   NEC Laboratories America and IDIAP Research Institute nor the names
   of its contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

Forked from PyTorch 2.0.1
"""

import builtins
import math


class SymInt:
    """
    Like an int (including magic methods), but redirects all operations on the
    wrapped node. This is used in particular to symbolically record operations
    in the symbolic shape workflow.
    """

    def __init__(self, node):
        # This field MUST be named node; C++ binding code assumes that this
        # class has a field named node that stores SymNode
        self.node = node

    def __bool__(self):
        return self.node.bool_()

    def __int__(self):
        return self.node.int_()

    # Magic methods installed by torch.fx.experimental.symbolic_shapes

    def __eq__(self, other: object) -> builtins.bool:
        raise AssertionError("type stub not overridden")

    def __lt__(self, other) -> builtins.bool:
        raise AssertionError("type stub not overridden")

    def __gt__(self, other) -> builtins.bool:
        raise AssertionError("type stub not overridden")

    def __le__(self, other) -> builtins.bool:
        raise AssertionError("type stub not overridden")

    def __ge__(self, other) -> builtins.bool:
        raise AssertionError("type stub not overridden")

    def __sym_max__(self, other):
        raise AssertionError("type stub not overridden")

    def __sym_min__(self, other):
        raise AssertionError("type stub not overridden")

    def __sym_float__(self):
        raise AssertionError("type stub not overridden")

    def __repr__(self):
        return str(self.node)


class SymFloat:
    """
    Like an float (including magic methods), but redirects all operations on the
    wrapped node. This is used in particular to symbolically record operations
    in the symbolic shape workflow.
    """

    def __init__(self, node):
        from torch.fx.experimental.symbolic_shapes import SymNode
        assert isinstance(node, SymNode)
        # This field MUST be named node; C++ binding code assumes that this
        # class has a field named node that stores SymNode
        self.node = node

    def __bool__(self):
        return self.node.bool_()

    # Magic methods installed by torch.fx.experimental.symbolic_shapes

    def __eq__(self, other: object) -> builtins.bool:
        raise AssertionError("type stub not overridden")

    def __lt__(self, other) -> builtins.bool:
        raise AssertionError("type stub not overridden")

    def __gt__(self, other) -> builtins.bool:
        raise AssertionError("type stub not overridden")

    def __le__(self, other) -> builtins.bool:
        raise AssertionError("type stub not overridden")

    def __ge__(self, other) -> builtins.bool:
        raise AssertionError("type stub not overridden")

    def __sym_max__(self, other):
        raise AssertionError("type stub not overridden")

    def __sym_min__(self, other):
        raise AssertionError("type stub not overridden")

    def __sym_int__(self):
        raise AssertionError("type stub not overridden")

    def __repr__(self):
        return self.node.str()


class SymBool:
    """
    Like an bool (including magic methods), but redirects all operations on the
    wrapped node. This is used in particular to symbolically record operations
    in the symbolic shape workflow.

    Unlike regular bools, regular boolean operators will force extra guards instead
    of symbolically evaluate.  Use the bitwise operators instead to handle this.
    """

    def __init__(self, node):
        from torch.fx.experimental.symbolic_shapes import SymNode
        assert isinstance(node, SymNode)
        # This field MUST be named node; C++ binding code assumes that this
        # class has a field named node that stores SymNode
        self.node = node

    def __bool__(self):
        return self.node.bool_()

    # Magic methods installed by torch.fx.experimental.symbolic_shapes
    def __and__(self, other) -> "SymBool":
        raise AssertionError("type stub not overridden")

    def __or__(self, other) -> "SymBool":
        raise AssertionError("type stub not overridden")

    # We very carefully define __sym_not__, and not a number of other
    # plausible alternatives:
    #
    #   - We do not override __not__ because this is not a real magic
    #     method; you cannot override the meaning of the not builtin in
    #     Python.  We use the name 'sym_not' to clarify that in user code you
    #     cannot use the builtin not or operator.not_ or operator.__not__ and
    #     hit this magic method; you must use our custom sym_not operator.
    #
    #   - We do not override the __invert__ method because SymBool is
    #     meant to be usable in situations where bool is expected.  However,
    #     bitwise negation ~a does the wrong thing with booleans (because
    #     bool is a subclass of int, so ~1 = -2 which is not falseish.)
    #     This would be a giant footgun, so we get around it by defining
    #     our own operator.  Note that bitwise and/or do the right thing,
    #     so we reuse the conventional operators there for readability.
    #
    def __sym_not__(self) -> "SymBool":
        raise AssertionError("type stub not overridden")

    def __repr__(self):
        return self.node.str()


def sym_not(a):
    r""" SymInt-aware utility for logical negation.

    Args:
        a (SymBool or bool): Object to negate
    """
    if hasattr(a, '__sym_not__'):
        return a.__sym_not__()
    return not a


def sym_float(a):
    r""" SymInt-aware utility for float casting.

    Args:
        a (SymInt, SymFloat, or object): Object to cast
    """
    if isinstance(a, SymFloat):
        return a
    elif hasattr(a, '__sym_float__'):
        return a.__sym_float__()
    return py_float(a)  # type: ignore[operator]


def sym_int(a):
    r""" SymInt-aware utility for int casting.

    Args:
        a (SymInt, SymFloat, or object): Object to cast
    """
    if isinstance(a, SymInt):
        return a
    elif isinstance(a, SymFloat):
        return math.floor(a) if a >= 0 else math.ceil(a)  # type: ignore[arg-type]
    return py_int(a)  # type: ignore[operator]


def sym_max(a, b):
    """ SymInt-aware utility for max()."""
    if isinstance(a, (SymInt, SymFloat)):
        return a.__sym_max__(b)
    elif isinstance(b, (SymInt, SymFloat)):
        # NB: If you actually care about preserving output type exactly
        # if you do something like max(0, 0.0), it is NOT sound to treat
        # min/max as commutative
        return b.__sym_max__(a)
    return builtins.max(a, b)  # type: ignore[operator]


def sym_min(a, b):
    """ SymInt-aware utility for max()."""
    if isinstance(a, (SymInt, SymFloat)):
        return a.__sym_min__(b)
    elif isinstance(b, (SymInt, SymFloat)):
        return b.__sym_min__(a)
    return builtins.min(a, b)  # type: ignore[operator]


# Populate magic methods on SymInt and SymFloat
import brevitas.backport.fx.experimental.symbolic_shapes
import brevitas.backport.fx



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import math
from typing import Optional, Tuple

import torch
from torch import Tensor
from torch.nn import Parameter

import brevitas
from brevitas import config
from brevitas.core.function_wrapper.misc import Identity
from brevitas.core.function_wrapper.ops_ste import ScalarClampMinSte
from brevitas.core.utils import StatelessBuffer
from brevitas.function.ops import max_int
from brevitas.quant_tensor import _unpack_quant_tensor
# Use custom implementation of kthvalue as work around to (b)float16 kernel limitations
from brevitas.utils.torch_utils import kthvalue

from .stats_wrapper import SCALAR_SHAPE

DEFAULT_STD_DEV_EPSILON = 1e-8


class NegativeMinOrZero(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim', 'keepdim']

    def __init__(
            self,
            stats_reduce_dim: Optional[int] = None,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None,
            keepdim: bool = False) -> None:
        super(NegativeMinOrZero, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim
        self.zero = StatelessBuffer(torch.tensor(0.0, dtype=dtype, device=device))
        self.keepdim = keepdim

    @brevitas.jit.script_method
    def forward(self, x: Tensor) -> Tensor:
        if self.stats_reduce_dim is None:
            min_val = torch.min(x)
        else:
            min_val = torch.min(x, dim=self.stats_reduce_dim, keepdim=self.keepdim)[0]
        min_val = torch.clamp(min_val, max=self.zero())
        return min_val


class AbsPercentile(brevitas.jit.ScriptModule):
    __constants__ = ['q', 'stats_reduce_dim', 'keepdim']

    def __init__(
            self,
            high_percentile_q: float,
            stats_reduce_dim: Optional[int],
            percentile_q=None,
            keepdim: bool = False):
        super(AbsPercentile, self).__init__()
        if percentile_q is not None:
            raise RuntimeError("percentile_q is deprecated, please pass high_percentile_q.")
        assert high_percentile_q <= 100, "q has to be a percentage"
        self.q = high_percentile_q
        self.stats_reduce_dim = stats_reduce_dim
        self.keepdim = keepdim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        if self.stats_reduce_dim is None:
            # k is 1-indexed, so round away from zero
            k = int(math.floor(.01 * self.q * x.numel() + 0.5))
            result = kthvalue(x.abs().view(-1), k)[0]
        else:
            # assuming x is two dimensional, get the other dimension
            assert len(x.size()) == 2, "Only 2-dim input is supported."
            other_dim = abs(self.stats_reduce_dim - 1)
            dim_slice = torch.narrow(x, dim=other_dim, start=0, length=1)
            # k is 1-indexed, so round away from zero
            k = int(math.floor(.01 * self.q * dim_slice.numel() + 0.5))
            result = kthvalue(x.abs(), k, dim=self.stats_reduce_dim, keepdim=self.keepdim)[0]
        return result


class NegativePercentileOrZero(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim', 'q', 'keepdim']

    def __init__(
            self,
            low_percentile_q,
            stats_reduce_dim: Optional[int] = None,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None,
            keepdim: bool = False) -> None:
        super(NegativePercentileOrZero, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim
        self.q = low_percentile_q
        self.zero = StatelessBuffer(torch.tensor(0.0, dtype=dtype, device=device))
        self.keepdim = keepdim

    @brevitas.jit.script_method
    def forward(self, x: Tensor) -> Tensor:
        if self.stats_reduce_dim is None:
            # k is 1-indexed, so round away from zero
            k = int(math.ceil(.01 * self.q * x.numel()))
            result = kthvalue(x.view(-1), k)[0]
        else:
            # assuming x is two dimensional, get the other dimension
            assert len(x.size()) == 2, "Only 2-dim input is supported."
            other_dim = abs(self.stats_reduce_dim - 1)
            dim_slice = torch.narrow(x, dim=other_dim, start=0, length=1)
            # k is 1-indexed, so round away from zero
            k = int(math.ceil(.01 * self.q * dim_slice.numel()))
            result = kthvalue(x, k, dim=self.stats_reduce_dim, keepdim=self.keepdim)[0]
        result = torch.clamp(result, max=self.zero())
        return result


class PercentileInterval(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim', 'low_q', 'high_q', 'keepdim']

    def __init__(
            self,
            low_percentile_q,
            high_percentile_q,
            stats_reduce_dim: Optional[int] = None,
            keepdim: bool = False,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None) -> None:
        super(PercentileInterval, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim
        self.low_q = low_percentile_q
        self.high_q = high_percentile_q
        self.keepdim = keepdim
        self.zero = StatelessBuffer(torch.tensor(0.0, dtype=dtype, device=device))

    @brevitas.jit.script_method
    def forward(self, x: Tensor) -> Tensor:
        if self.stats_reduce_dim is None:
            low_k = int(math.ceil(.01 * self.low_q * x.numel()))
            # k is 1-indexed, so round away from zero
            high_k = int(math.floor(.01 * self.high_q * x.numel() + 0.5))
            low_result = kthvalue(x.view(-1), low_k)[0]
            high_result = kthvalue(x.view(-1), high_k)[0]
        else:
            # assuming x is two dimensional, get the other dimension
            assert len(x.size()) == 2, "Only 2-dim input is supported."
            other_dim = abs(self.stats_reduce_dim - 1)
            dim_slice = torch.narrow(x, dim=other_dim, start=0, length=1)
            low_k = int(math.ceil(.01 * self.low_q * dim_slice.numel()))
            # k is 1-indexed, so round away from zero
            high_k = int(math.floor(.01 * self.high_q * dim_slice.numel() + 0.5))
            low_result = kthvalue(x, low_k, dim=self.stats_reduce_dim, keepdim=self.keepdim)[0]
            high_result = kthvalue(x, high_k, dim=self.stats_reduce_dim, keepdim=self.keepdim)[0]
        # We need to make sure the lower bound is not positive to align with zero-point statistics
        low_result = torch.clamp(low_result, max=self.zero())
        interval = high_result - low_result
        abs_interval = torch.abs(interval)
        return abs_interval


class AbsMax(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: Optional[int] = None, keepdim: bool = False) -> None:
        super(AbsMax, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim
        self.keepdim = keepdim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        if self.stats_reduce_dim is None:
            return torch.max(torch.abs(x))
        else:
            return torch.max(torch.abs(x), dim=self.stats_reduce_dim, keepdim=self.keepdim)[0]


class AbsMinMax(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim', 'keepdim']

    def __init__(
            self,
            stats_reduce_dim: Optional[int] = None,
            keepdim: bool = False,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None) -> None:
        super(AbsMinMax, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim
        self.keepdim = keepdim
        self.zero = StatelessBuffer(torch.tensor(0.0, dtype=dtype, device=device))

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        if self.stats_reduce_dim is None:
            max_val = torch.max(x)
            min_val = torch.min(x)
        else:
            max_val = torch.max(x, dim=self.stats_reduce_dim, keepdim=self.keepdim)[0]
            min_val = torch.min(x, dim=self.stats_reduce_dim, keepdim=self.keepdim)[0]
        # We need to make sure the lower bound is not positive to align with zero-point statistics
        min_val = torch.clamp(min_val, max=self.zero())
        return torch.abs(max_val - min_val)


class AbsMaxAve(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: int) -> None:
        super(AbsMaxAve, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        return torch.mean(torch.max(torch.abs(x), dim=self.stats_reduce_dim)[0])


class AbsMaxL2(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: int) -> None:
        super(AbsMaxL2, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor):
        per_channel_max = torch.max(torch.abs(x), dim=self.stats_reduce_dim)[0]
        out = torch.norm(per_channel_max, p=2)
        out = out / math.sqrt(per_channel_max.view(-1).shape[0])
        return out


class AbsAve(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: Optional[int] = None) -> None:
        super(AbsAve, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        if self.stats_reduce_dim is None:
            return torch.mean(torch.abs(x))
        else:
            return torch.mean(torch.abs(x), dim=self.stats_reduce_dim)


class MeanSigmaStd(brevitas.jit.ScriptModule):

    def __init__(
            self,
            sigma: float,
            stats_reduce_dim: Optional[int] = None,
            std_dev_epsilon: float = DEFAULT_STD_DEV_EPSILON,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None) -> None:
        super(MeanSigmaStd, self).__init__()
        self.impl = _MeanSigmaStdImpl(stats_reduce_dim, std_dev_epsilon)
        self.sigma = StatelessBuffer(torch.tensor(sigma, dtype=dtype, device=device))

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        sigma = self.sigma()
        out = self.impl(x, sigma)
        return out


class _MeanSigmaStdImpl(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim', 'output_shape', 'epsilon']

    def __init__(
            self,
            stats_reduce_dim: Optional[int] = None,
            std_dev_epsilon: float = DEFAULT_STD_DEV_EPSILON) -> None:
        super(_MeanSigmaStdImpl, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim
        self.epsilon = std_dev_epsilon

    @brevitas.jit.script_method
    def forward(self, x: Tensor, sigma: Tensor):
        abs_val = torch.abs(x)
        if self.stats_reduce_dim is None:
            mean_val = torch.mean(abs_val)
            std_val = torch.sqrt(torch.var(abs_val) + self.epsilon)
        else:
            mean_val = torch.mean(torch.abs(x), dim=self.stats_reduce_dim)
            std_val = torch.sqrt(torch.var(abs_val, dim=self.stats_reduce_dim) + self.epsilon)
            mean_val = mean_val.view(-1)
            std_val = std_val.view(-1)
        return mean_val + sigma * std_val


class MeanLearnedSigmaStd(brevitas.jit.ScriptModule):

    def __init__(
            self,
            sigma: float,
            stats_output_shape: Tuple[int, ...],
            stats_reduce_dim: Optional[int] = None,
            std_dev_epsilon: float = DEFAULT_STD_DEV_EPSILON,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None) -> None:
        super(MeanLearnedSigmaStd, self).__init__()
        self.impl = _MeanSigmaStdImpl(stats_reduce_dim, std_dev_epsilon)
        if stats_output_shape == SCALAR_SHAPE:
            self.value = Parameter(torch.tensor(sigma, dtype=dtype, device=device))
        else:
            self.value = Parameter(
                torch.full(stats_output_shape, sigma, dtype=dtype, device=device))

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        sigma = self.sigma.view(self.sigma.shape)  # trick to get a tensor type
        out = self.impl(x, sigma)
        return out

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        value_key = prefix + 'sigma'
        retrocomp_value_key = prefix + 'learned_sigma'
        if retrocomp_value_key in state_dict:  # retrocompatibility
            state_dict[value_key] = state_dict.pop(retrocomp_value_key)
        super(MeanLearnedSigmaStd, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        sigma_key = prefix + 'sigma'
        if config.IGNORE_MISSING_KEYS and sigma_key in missing_keys:
            missing_keys.remove(sigma_key)


class KLMinimizerThreshold(torch.nn.Module):
    """
    Based on:
    https://github.com/apache/incubator-mxnet/blob/master/python/mxnet/contrib/quantization.py
    """

    def __init__(self, signed, bit_width_impl, num_bins=1000 + 1, smoothing_eps=0.0001):
        super(KLMinimizerThreshold, self).__init__()
        self.num_bins = num_bins
        self.smoothing_eps = smoothing_eps
        self.signed = signed
        self.bit_width_impl = bit_width_impl
        self.absmax_impl = AbsMax()

    def smooth_normalize_distribution(self, p, eps):
        is_zeros = (p == 0).float()
        n_zeros = is_zeros.sum()
        n_nonzeros = torch.numel(p) - n_zeros
        if not n_nonzeros:
            return None
        eps1 = eps * n_zeros / n_nonzeros
        hist = p.float()
        hist += eps * is_zeros + (-eps1) * n_nonzeros
        dist = torch.distributions.categorical.Categorical(logits=hist)
        return dist

    def forward(self, x: Tensor):
        absmax = self.absmax_impl(x)
        bit_width = self.bit_width_impl()
        num_quantized_bins = max_int(self.signed, False, bit_width).int()
        thresholds = torch.zeros(self.num_bins // 2 + 1 - num_quantized_bins // 2, device=x.device)
        divergence = torch.zeros_like(thresholds)
        quantized_bins = torch.zeros(num_quantized_bins, device=x.device)
        hist = torch.histc(x, bins=self.num_bins, min=-absmax, max=absmax).int()
        hist_edges = torch.linspace(-absmax, absmax, self.num_bins + 1)
        for i in range(num_quantized_bins // 2, self.num_bins // 2 + 1):
            p_bin_idx_start = self.num_bins // 2 - i
            p_bin_idx_stop = self.num_bins // 2 + i + 1
            thresholds[i - num_quantized_bins // 2] = hist_edges[p_bin_idx_stop]
            sliced_nd_hist = hist[p_bin_idx_start:p_bin_idx_stop]
            p = sliced_nd_hist.clone()
            left_outlier_count = torch.sum(hist[0:p_bin_idx_start])
            p[0] += left_outlier_count
            right_outlier_count = torch.sum(hist[p_bin_idx_stop:])
            p[-1] += right_outlier_count
            is_nonzeros = (sliced_nd_hist != 0).float()
            num_merged_bins = torch.numel(p) // num_quantized_bins
            for j in range(num_quantized_bins):
                start = j * num_merged_bins
                stop = start + num_merged_bins
                quantized_bins[j] = sliced_nd_hist[start:stop].sum()
            quantized_bins[-1] += sliced_nd_hist[num_quantized_bins * num_merged_bins:].sum()
            q = torch.zeros_like(p, dtype=torch.float32, device=x.device)
            for j in range(num_quantized_bins):
                start = j * num_merged_bins
                if j == num_quantized_bins - 1:
                    stop = -1
                else:
                    stop = start + num_merged_bins
                norm = is_nonzeros[start:stop].sum()
                if norm != 0:
                    q[start:stop] = quantized_bins[j] / norm
            q[sliced_nd_hist == 0] = 0.
            p = self.smooth_normalize_distribution(p, self.smoothing_eps)
            q = self.smooth_normalize_distribution(q, self.smoothing_eps)
            if q is None:
                divergence[i - num_quantized_bins // 2] = float('inf')
            else:
                divergence[i - num_quantized_bins // 2] = torch.distributions.kl.kl_divergence(p, q)
        min_divergence_idx = torch.argmin(divergence)
        opt_threshold = thresholds[min_divergence_idx]
        return opt_threshold


class L1Norm(brevitas.jit.ScriptModule):
    """ScriptModule implementation to collect per-channel L1 normalization stats
    for weight normalization-based quantization."""
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: Optional[int] = None) -> None:
        super(L1Norm, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        if self.stats_reduce_dim is None:
            # Need to be able to return the max per-channel L1 norm as a scalar
            raise NotImplementedError("L1 normalization is not supported per-tensor yet.")
        else:
            return x.norm(p=1, dim=self.stats_reduce_dim, keepdim=True)


class L2Norm(brevitas.jit.ScriptModule):
    """ScriptModule implementation to collect per-channel L2 normalization stats
    for weight normalization-based quantization."""
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: Optional[int] = None) -> None:
        super(L2Norm, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        if self.stats_reduce_dim is None:
            # Need to be able to return the max per-channel L2 norm as a scalar
            raise NotImplementedError("L2 normalization is not supported per-tensor yet.")
        else:
            return x.norm(p=2, dim=self.stats_reduce_dim, keepdim=True)


def _set_local_loss_mode(module, enabled):
    for m in module.modules():
        if hasattr(m, 'local_loss_mode'):
            m.local_loss_mode = enabled


def _set_observer_mode(module, enabled, previous_observer_mode):
    for m in module.modules():
        if hasattr(m, 'observer_only'):
            previous_observer_mode[m] = m.observer_only
            m.observer_only = enabled


def _restore_observer_mode(module, previous_observer_mode):
    for m in module.modules():
        if hasattr(m, 'observer_only'):
            m.observer_only = previous_observer_mode[m]


class MSE(torch.nn.Module):
    # References:
    # https://github.com/cornell-zhang/dnn-quant-ocs/blob/master/distiller/quantization/clip.py
    # https://github.com/wimh966/outlier_suppression/blob/main/quant_transformer/quantization/observer.py

    def __init__(
            self,
            proxy_module,
            mse_init_op,
            inner_stats_input_view_shape_impl: torch.nn.Module,
            stats_reduce_dim: Optional[int] = None,
            mse_search_method: str = 'fibonacci',
            mse_iters: int = 20):
        super(MSE, self).__init__()
        self.mse_init_op = mse_init_op
        self.input_view_shape_impl = inner_stats_input_view_shape_impl
        self.proxy_forward = proxy_module.forward
        self.previous_observer_mode = dict()
        self.set_local_loss_mode = lambda enabled: _set_local_loss_mode(proxy_module, enabled)
        self.set_observer_mode = lambda enabled: _set_observer_mode(
            proxy_module, enabled, self.previous_observer_mode)
        self.restore_observer_mode = lambda: _restore_observer_mode(
            proxy_module, self.previous_observer_mode)
        self.internal_candidate = None
        self.num = mse_iters
        self.search_method = mse_search_method
        self.stats_reduce_dim = stats_reduce_dim
        self.local_loss_mode: bool = False

    def mse_loss_fn(self, x, quant_value):
        loss = torch.nn.functional.mse_loss(x, quant_value, reduction='none')
        if self.stats_reduce_dim is not None:
            # stats_reduce_dim applies to the permuted and reshaped tensor
            loss = self.input_view_shape_impl(loss)
            loss = torch.sum(loss, dim=self.stats_reduce_dim)
        else:
            loss = torch.sum(loss)
        return loss

    def evaluate_loss(self, x, candidate):
        self.internal_candidate = candidate
        # Set to local_loss_mode before calling the proxy
        self.set_local_loss_mode(True)
        self.set_observer_mode(False)
        quant_value = self.proxy_forward(x)
        quant_value = _unpack_quant_tensor(quant_value)
        loss = self.mse_loss_fn(x, quant_value)
        self.set_local_loss_mode(False)
        self.restore_observer_mode()
        return loss

    def mse_grid_search(self, xl, x):
        best_loss = torch.tensor(float('inf'), device=x.device, dtype=x.dtype)
        best_candidate = xl
        for i in range(2, self.num + 1):
            candidate = (xl * i).detach()
            loss = self.evaluate_loss(x, candidate)
            best_candidate = torch.where(loss < best_loss, candidate, best_candidate)
            best_loss = torch.min(loss, best_loss)
        return best_candidate

    def mse_fib_search(self, xl, xr, x):

        def fib_seq(n):
            if n <= 0:
                return [0]
            seq = [0, 1]
            while len(seq) <= n:
                next = seq[-1] + seq[-2]
                seq.append(next)
            return seq

        # vectorized variant of
        # https://indrag49.github.io/Numerical-Optimization/solving-one-dimensional-optimization-problems.html#fibonacci-search-method
        F = fib_seq(self.num)
        L0 = xr - xl
        Li = (F[self.num - 2] / F[self.num]) * L0
        for i in range(2, self.num + 1):
            x1 = torch.where(Li > L0 / 2, xr - Li, xl + Li)
            x2 = torch.where(Li > L0 / 2, xl + Li, xr - Li)
            f1, f2 = self.evaluate_loss(x, x1), self.evaluate_loss(x, x2)
            xr = torch.where(f1 <= f2, x2, xr)
            xl = torch.where(f1 >= f2, x1, xl)
            Li = (F[self.num - i] / F[self.num - (i - 2)]) * torch.where(f1 != f2, L0, xr - xl)
            L0 = xr - xl
        return torch.where(f1 <= f2, x1, x2)

    def mse_search(self, x):
        x_view = self.input_view_shape_impl(x)
        init = self.mse_init_op(x_view).detach()
        base = init / self.num
        if self.search_method == 'grid':
            best_candidate = self.mse_grid_search(base, x)
        elif self.search_method == 'fibonacci':
            best_candidate = self.mse_fib_search(base, init, x)
        else:
            raise ValueError(f"Search method {self.search_method} not supported.")
        # Save for evaluation by other modules (e.g. zp) invoking local loss mode
        self.internal_candidate = best_candidate
        return best_candidate

    def forward(self, x):
        if not self.local_loss_mode:
            with torch.no_grad():
                return self.mse_search(x)
        else:
            # This is invoked for the zero-point whenever scale is being optimized first
            if self.internal_candidate is None:
                x = self.input_view_shape_impl(x)
                self.internal_candidate = self.mse_init_op(x).detach()
            return self.internal_candidate


class HalfQuadraticOptimizerScale(torch.nn.Module):
    # References:
    # https://mobiusml.github.io/hqq_blog/
    # https://github.com/mobiusml/hqq?tab=readme-ov-file

    def __init__(
            self,
            proxy_module,
            hqo_init_op_scale,
            keepdim: bool,
            inner_stats_input_view_shape_impl: torch.nn.Module,
            scaling_min_val: Optional[float] = None,
            stats_reduce_dim: Optional[int] = None,
            int_scaling_impl=None,
            bit_width_impl=None,
            hqo_beta_scale: float = 1e5,
            hqo_kappa_scale: float = 1.01,
            hqo_lp_norm_scale: float = .7,
            hqo_iters_scale: int = 1000):
        super(HalfQuadraticOptimizerScale, self).__init__()
        self.hqo_init_op = hqo_init_op_scale
        self.input_view_shape_impl = inner_stats_input_view_shape_impl
        self.proxy_forward = proxy_module.forward
        self.previous_observer_mode = dict()
        self.set_local_loss_mode = lambda enabled: _set_local_loss_mode(proxy_module, enabled)
        self.set_observer_mode = lambda enabled: _set_observer_mode(
            proxy_module, enabled, self.previous_observer_mode)
        self.restore_observer_mode = lambda: _restore_observer_mode(
            proxy_module, self.previous_observer_mode)
        self.internal_candidate = None
        self.hqo_iters = hqo_iters_scale
        self.stats_reduce_dim = stats_reduce_dim
        self.local_loss_mode: bool = False

        self.beta = hqo_beta_scale
        self.kappa = hqo_kappa_scale
        self.lp_norm = hqo_lp_norm_scale

        self.int_scaling_impl = int_scaling_impl
        self.msb_clamp_bit_width_impl = bit_width_impl
        if scaling_min_val is not None and scaling_min_val != 0:
            self.clamp_min_ste = ScalarClampMinSte(scaling_min_val)
        else:
            self.clamp_min_ste = Identity()
        self.keepdim = keepdim

    def parameter_search(self, xl, x):
        best_loss = torch.tensor(float('inf'), device=x.device, dtype=x.dtype)
        candidate = xl
        best_candidate = candidate
        beta = self.beta
        with torch.no_grad():
            for i in range(0, self.hqo_iters):
                self.internal_candidate = candidate
                self.set_local_loss_mode(True)
                self.set_observer_mode(False)
                quant_tensor = self.proxy_forward(x).detach()
                self.set_local_loss_mode(False)
                self.restore_observer_mode()
                loss = torch.abs(quant_tensor.value - x).mean()

                best_candidate = torch.where(loss < best_loss, candidate, best_candidate)
                if loss >= best_loss:
                    break
                best_loss = torch.min(loss, best_loss)
                W_e = shrink_lp_op(x - quant_tensor.value, beta, self.lp_norm)
                zero_point = quant_tensor.zero_point
                num = self.input_view_shape_impl(x - W_e).detach()
                den = self.input_view_shape_impl(
                    torch.round(quant_tensor.value / quant_tensor.scale) - zero_point).detach()
                mask = (num != 0.) & (den != 0.)
                if self.stats_reduce_dim is None:
                    candidate = masked_median(num / den, mask)
                else:
                    candidate = masked_median(
                        num / den, mask, dim=self.stats_reduce_dim, keepdim=self.keepdim)
                candidate = candidate.type_as(self.internal_candidate)
                candidate = self.clamp_min_ste(candidate)
                bit_width = self.msb_clamp_bit_width_impl()
                int_threshold = self.int_scaling_impl(bit_width)
                candidate = candidate * int_threshold
                candidate[torch.isnan(candidate)] = self.internal_candidate[torch.isnan(candidate)]
                candidate[torch.isinf(candidate)] = self.internal_candidate[torch.isinf(candidate)]
                beta *= self.kappa
        return best_candidate

    def optimize(self, x):
        x_view = self.input_view_shape_impl(x)

        init = self.hqo_init_op(x_view).detach()
        best_candidate = self.parameter_search(init, x_view)

        # Save for evaluation by other modules (e.g. zp) invoking local loss mode
        self.internal_candidate = best_candidate.detach()
        torch.cuda.empty_cache()
        return best_candidate

    def forward(self, x):
        if not self.local_loss_mode:
            with torch.no_grad():
                return self.optimize(x)
        else:
            # This is invoked for the zero-point whenever scale is being optimized first
            if self.internal_candidate is None:
                x = self.input_view_shape_impl(x)
                self.internal_candidate = self.hqo_init_op(x).detach()
            return self.internal_candidate


class HalfQuadraticOptimizerZeroPoint(torch.nn.Module):
    # References:
    # https://mobiusml.github.io/hqq_blog/
    # https://github.com/mobiusml/hqq?tab=readme-ov-file

    def __init__(
            self,
            proxy_module,
            keepdim: bool,
            hqo_init_op_zp: torch.nn.Module,
            inner_stats_input_view_shape_impl: torch.nn.Module,
            stats_reduce_dim: Optional[int] = None,
            hqo_beta_zp: float = 1e0,
            hqo_kappa_zp: float = 1.01,
            hqo_lp_norm_zp: float = .5,
            hqo_iters_zp: int = 1000):
        super(HalfQuadraticOptimizerZeroPoint, self).__init__()
        self.hqo_init_op_zp = hqo_init_op_zp
        self.input_view_shape_impl = inner_stats_input_view_shape_impl
        self.proxy_forward = proxy_module.forward
        self.previous_observer_mode = dict()
        self.set_local_loss_mode = lambda enabled: _set_local_loss_mode(proxy_module, enabled)
        self.set_observer_mode = lambda enabled: _set_observer_mode(
            proxy_module, enabled, self.previous_observer_mode)
        self.restore_observer_mode = lambda: _restore_observer_mode(
            proxy_module, self.previous_observer_mode)
        self.internal_candidate = None
        self.stats_reduce_dim = stats_reduce_dim
        self.local_loss_mode: bool = False
        self.beta = hqo_beta_zp
        self.kappa = hqo_kappa_zp
        self.lp_norm = hqo_lp_norm_zp
        self.hqo_iters = hqo_iters_zp
        self.keepdim = keepdim

    def parameter_search(self, xl, x):
        best_loss = torch.tensor(float('inf'), device=x.device, dtype=x.dtype)
        candidate = xl
        best_candidate = candidate
        with torch.no_grad():
            for i in range(0, self.hqo_iters):
                self.internal_candidate = candidate
                self.set_local_loss_mode(True)
                self.set_observer_mode(False)
                quant_tensor = self.proxy_forward(x).detach()
                self.set_local_loss_mode(False)
                self.restore_observer_mode()
                qt_value = self.input_view_shape_impl(quant_tensor.value)
                qt_scale = self.input_view_shape_impl(quant_tensor.scale)
                qt_zp = self.input_view_shape_impl(quant_tensor.zero_point)
                qt_int = qt_value / qt_scale + qt_zp
                loss = torch.abs(qt_value - x).mean()
                best_candidate = torch.where(loss < best_loss, candidate, best_candidate)
                if loss >= best_loss:
                    break
                best_loss = torch.min(loss, best_loss)
                W_e = shrink_lp_op(x - qt_value, self.beta, self.lp_norm)

                # Compared to the original formulation, the value we're looking for is:
                # - scaled by qt_scale
                # - opposite sign
                val = self.input_view_shape_impl((x - W_e) - qt_int * qt_scale)

                if self.stats_reduce_dim is None:
                    candidate = torch.mean(val)
                else:
                    candidate = torch.mean(val, dim=self.stats_reduce_dim, keepdim=self.keepdim)
                self.beta *= self.kappa
        return best_candidate

    def optimize(self, x):
        x_view = self.input_view_shape_impl(x)

        init = self.hqo_init_op_zp(x_view).detach()

        best_candidate = self.parameter_search(init, x)

        # Save for evaluation by other modules (e.g. zp) invoking local loss mode
        self.internal_candidate = best_candidate.detach()
        torch.cuda.empty_cache()
        return best_candidate

    def forward(self, x):
        if not self.local_loss_mode:
            with torch.no_grad():
                return self.optimize(x)
        else:
            # This is invoked for the zero-point whenever scale is being optimized first
            if self.internal_candidate is None:
                x = self.input_view_shape_impl(x)
                self.internal_candidate = self.hqo_init_op_zp(x).detach()
            return self.internal_candidate


def masked_median(x, mask, dim=None, keepdim=False):
    """Compute the median of tensor x along dim, ignoring values where mask is False.
    x and mask need to be broadcastable.

    Args:
        x (Tensor): Tensor to compute median of.
        mask (BoolTensor): Same shape as x with True where x is valid and False
            where x should be masked. Mask should not be all False in any column of
            dimension dim to avoid NaNs from zero division.
        dim (int, optional): Dimension to take median of. Defaults to 0.

    Returns:
        Tensor: Same shape as x, except dimension dim reduced.
    """
    # uncomment this assert for safety but might impact performance
    # assert (
    #     mask.sum(dim=dim).ne(0).all()
    # ), "mask should not be all False in any column, causes zero division"
    x_nan = x.float().masked_fill(~mask, float("nan"))
    if dim is None:
        x_median = x_nan.nanmedian()
    else:
        x_median, _ = x_nan.nanmedian(dim=dim, keepdim=keepdim)
    return x_median


# Shrinking operator
def shrink_lp_op(x: Tensor, beta: float, lp_norm: float) -> Tensor:
    if lp_norm == 1:
        return torch.sign(x) * torch.nn.functional.relu(torch.abs(x) - 1.0 / beta)
    else:
        return torch.sign(x) * torch.nn.functional.relu(
            torch.abs(x) - (1.0 / beta) * torch.pow(torch.abs(x), lp_norm - 1))



"""
Copyright (c) 2023-     Advanced Micro Devices, Inc. (Ian Colbert)
Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
Copyright (c) 2011-2013 NYU                      (Clement Farabet)
Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. Neither the names of AMD, Facebook, Deepmind Technologies, NYU,
   NEC Laboratories America and IDIAP Research Institute nor the names
   of its contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

import os
import tarfile
from typing import Tuple, Type

from PIL import Image
from six.moves import urllib
from torch.utils.data import DataLoader
import torch.utils.data as data
from torchvision.transforms import CenterCrop
from torchvision.transforms import Compose
from torchvision.transforms import RandomCrop
from torchvision.transforms import RandomHorizontalFlip
from torchvision.transforms import RandomVerticalFlip
from torchvision.transforms import Resize
from torchvision.transforms import ToTensor

__all__ = ["get_bsd300_dataloaders"]

url = "http://www2.eecs.berkeley.edu/Research/Projects/CS/vision/bsds/BSDS300-images.tgz"


def is_valid_image_file(filename: str):
    extensions = [".png", ".jpg", ".jpeg"]
    for extension in extensions:
        if filename.endswith(extension):
            return True
    return False


def load_img_ycbcr(filepath):
    img = Image.open(filepath).convert('YCbCr')
    y, _, _ = img.split()
    return y


def load_img_rbg(filepath):
    img = Image.open(filepath).convert('RGB')
    return img


class DatasetFromFolder(data.Dataset):

    def __init__(self, image_dir, shared_transform, input_transform, target_transform):
        super(DatasetFromFolder, self).__init__()
        self.image_filenames = [
            os.path.join(image_dir, x) for x in os.listdir(image_dir) if is_valid_image_file(x)]

        self.shared_transform = shared_transform
        self.input_transform = input_transform
        self.target_transform = target_transform

    def __getitem__(self, index):
        input = load_img_rbg(self.image_filenames[index])
        input = self.shared_transform(input)
        target = input.copy()
        input = self.input_transform(input)
        target = self.target_transform(target)
        return input, target

    def __len__(self):
        return len(self.image_filenames)


def download_bsd300(dest: str = "data", download: bool = False):
    if not isinstance(dest, str):
        raise ValueError("Specify dataset directory with --data-dir")
    output_image_dir = os.path.join(dest, "BSDS300/images")
    if not os.path.exists(output_image_dir) and download:
        os.makedirs(dest, exist_ok=True)
        print("Data does not exist. Downloading from ", url)
        data = urllib.request.urlopen(url)
        file_path = os.path.join(dest, os.path.basename(url))
        with open(file_path, 'wb') as f:
            f.write(data.read())
        with tarfile.open(file_path) as tar:
            for item in tar:
                tar.extract(item, dest)
        os.remove(file_path)
    return output_image_dir


def calculate_valid_crop_size(crop_size, upscale_factor):
    return crop_size - (crop_size % upscale_factor)


def train_transforms(crop_size):
    return Compose([
        RandomCrop(crop_size, pad_if_needed=True), RandomHorizontalFlip(), RandomVerticalFlip()])


def test_transforms(crop_size):
    return Compose([CenterCrop(crop_size)])


def input_transform(crop_size, upscale_factor):
    return Compose([
        Resize(crop_size // upscale_factor),
        ToTensor(),])


def target_transform():
    return Compose([ToTensor()])


def get_training_set(upscale_factor: int, root_dir: str, crop_size: int):
    train_dir = os.path.join(root_dir, "train")
    crop_size = calculate_valid_crop_size(crop_size, upscale_factor)
    return DatasetFromFolder(
        train_dir,
        shared_transform=train_transforms(crop_size),
        input_transform=input_transform(crop_size, upscale_factor),
        target_transform=target_transform())


def get_test_set(upscale_factor: int, root_dir: str, crop_size: int):
    test_dir = os.path.join(root_dir, "test")
    crop_size = calculate_valid_crop_size(crop_size, upscale_factor)
    return DatasetFromFolder(
        test_dir,
        shared_transform=test_transforms(crop_size),
        input_transform=input_transform(crop_size, upscale_factor),
        target_transform=target_transform())


def get_bsd300_dataloaders(
        data_root: str,
        num_workers: int = 0,
        batch_size: int = 32,
        batch_size_test: int = 100,
        pin_memory: bool = True,
        upscale_factor: int = 3,
        crop_size: int = 512,
        download: bool = False) -> Tuple[Type[DataLoader]]:
    """Function that loads BSD300 dataset from data_root folder and returns the training
    and testing dataloaders. If <data_root>/BSD300/images does not exist, then the data is
    downloaded.

    Args:
        data_root (str): Root folder containing the BSD300 dataset.
        num_workers (int): Number of workers to use for both dataloaders. Default: 0
        batch_size (int): Size of batches to use for the training dataloader. Default: 32
        batch_size_test (int): Size of batches to use for the testing dataloader. When
            None, then batch_size_test = batch_size. Default: 100
        pin_memory (bool): Whether or not to pin the memory for both dataloaders. Default: True
        upscale_factor (int): The upscale factor for the super resolution task. Default: 3
        crop_size (int): The size to crop images for upscaling. Default: 512
        download (bool): Whether or not to download the dataset. Default: False
    """
    data_root = download_bsd300(data_root, download)
    trainset = get_training_set(upscale_factor, data_root, crop_size)
    trainloader = DataLoader(
        trainset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=num_workers,
        pin_memory=pin_memory,
        persistent_workers=num_workers > 0)
    testset = get_test_set(upscale_factor, data_root, crop_size)
    if batch_size_test is None:
        batch_size_test = batch_size
    testloader = DataLoader(
        testset,
        batch_size=batch_size_test,
        shuffle=False,
        num_workers=num_workers,
        pin_memory=pin_memory,
        persistent_workers=num_workers > 0)
    return trainloader, testloader



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from abc import abstractmethod

import torch
from torch import Tensor

from brevitas.export.common.handler.base import BaseHandler
from brevitas.export.onnx import onnx_export_opset
from brevitas.export.onnx.debug import DebugMarkerFunction
from brevitas.nn.quant_rnn import _QuantLSTMLayer

from .function import LSTMCellFn

__all__ = [
    'Kernel1dApplHandlerMixin',
    'Kernel2dApplHandlerMixin',
    'ONNXBaseHandler',
    'QuantLSTMLayerHandler']


class Kernel1dApplHandlerMixin(ABC):

    @staticmethod
    def padding(module):
        if isinstance(module.padding, int):
            padding = [module.padding] * 2
        else:
            padding = list(module.padding)
            if len(padding) == 1:
                return padding + padding
        return padding

    @staticmethod
    def stride(module):
        if isinstance(module.stride, int):
            return [module.stride]
        else:
            return list(module.stride)

    @staticmethod
    def dilation(module):
        if isinstance(module.dilation, int):
            return [module.dilation]
        else:
            dilation = list(module.dilation)
            return dilation

    @staticmethod
    def kernel_shape(module):
        if isinstance(module.kernel_size, int):
            return [module.kernel_size]
        else:
            return list(module.kernel_size)


class Kernel2dApplHandlerMixin(ABC):

    @staticmethod
    def padding(module):
        if isinstance(module.padding, int):
            padding = [module.padding] * 4
        else:
            padding = list(module.padding) + list(module.padding)
        return padding

    @staticmethod
    def stride(module):
        if isinstance(module.stride, int):
            return [module.stride] * 2
        else:
            return list(module.stride)

    @staticmethod
    def dilation(module):
        if isinstance(module.dilation, int):
            return [module.dilation] * 2
        else:
            return list(module.dilation)

    @staticmethod
    def kernel_shape(module):
        if isinstance(module.kernel_size, int):
            return [module.kernel_size] * 2
        else:
            return list(module.kernel_size)


class Kernel3dApplHandlerMixin(ABC):

    @staticmethod
    def padding(module):
        if isinstance(module.padding, int):
            padding = [module.padding] * 6
        else:
            padding = list(module.padding) + list(module.padding)
        return padding

    @staticmethod
    def stride(module):
        if isinstance(module.stride, int):
            return [module.stride] * 3
        else:
            return list(module.stride)

    @staticmethod
    def dilation(module):
        if isinstance(module.dilation, int):
            return [module.dilation] * 3
        else:
            return list(module.dilation)

    @staticmethod
    def kernel_shape(module):
        if isinstance(module.kernel_size, int):
            return [module.kernel_size] * 3
        else:
            return list(module.kernel_size)


class ONNXBaseHandler(BaseHandler, ABC):

    def __init__(self):
        super().__init__()
        self.symbolic_kwargs = {}
        self.export_debug_name = None
        self.debug_input = False
        self.debug_output = False

    @abstractmethod
    def symbolic_execution(self, *args, **kwargs):
        pass

    def attach_debug_info(self, m):
        self.export_debug_name = m.export_debug_name
        self.debug_input = m.export_input_debug
        self.debug_output = m.export_output_debug

    def forward(self, inp: Tensor, *args, **kwargs):
        debug_fn = lambda x, name: DebugMarkerFunction.apply(x, self.export_debug_name + name)
        if self.export_debug_name is not None and self.debug_input:
            inp = debug_fn(inp, '.input')
        out = self.symbolic_execution(inp, *args, **kwargs)
        if self.export_debug_name is not None and self.debug_output:
            if isinstance(out, Tensor):
                out = debug_fn(out, '.output')
            elif isinstance(out, tuple) and isinstance(out[0], Tensor):
                out = list(out)
                out[0] = debug_fn(out[0], '.output')
                out = tuple(out)
        return out


class QuantLSTMLayerHandler(ONNXBaseHandler, ABC):
    handled_layer = _QuantLSTMLayer

    def prepare_for_export(self, module: _QuantLSTMLayer):
        self.symbolic_kwargs = {
            'batch_first': module.cell.batch_first,
            'reverse_input': module.cell.reverse_input,
            'cifg': module.cifg}
        quantizers = [
            'output',
            'cell_state',
            'input_acc',
            'forget_acc',
            'cell_acc',
            'output_acc',
            'input_sigmoid',
            'forget_sigmoid',
            'cell_tanh',
            'output_sigmoid',
            'hidden_state_tanh']
        quantizers = {name: getattr(module.cell, name + '_quant') for name in quantizers}

        if all([q.is_quant_enabled for q in quantizers.values()]):
            self.quantized_cell = True
        elif all([not q.is_quant_enabled for q in quantizers.values()]):
            self.quantized_cell = False
        else:
            raise RuntimeError("Export of a partially quantized LSTM cell not supported.")

        if self.quantized_cell:
            for (name, quant) in quantizers.items():
                # ONNX export doesn't handle optional values well, so in case of cifg
                # we duplicate the forget ones to the input ones
                if module.cifg and name == 'input_acc':
                    quant = quantizers['forget_acc']
                elif module.cifg and name == 'input_sigmoid':
                    quant = quantizers['forget_sigmoid']
                self.symbolic_kwargs[name + '_scale'] = quant.scale()
                self.symbolic_kwargs[name + '_zero_point'] = quant.zero_point()
                self.symbolic_kwargs[name + '_bit_width'] = quant.bit_width()
                self.symbolic_kwargs[name + '_narrow_range'] = quant.is_narrow_range
                self.symbolic_kwargs[name + '_signed'] = quant.is_signed
                self.symbolic_kwargs[name + '_rounding_mode'] = quant.rounding_mode

    @abstractmethod
    def quantized_cell_symbolic_execution(
            self,
            quant_input,
            quant_hidden_state,
            quant_cell_state,
            quant_weight_ii,
            quant_weight_if,
            quant_weight_ic,
            quant_weight_io,
            quant_weight_hi,
            quant_weight_hf,
            quant_weight_hc,
            quant_weight_ho,
            quant_bias_input,
            quant_bias_forget,
            quant_bias_cell,
            quant_bias_output):
        pass

    def symbolic_execution(
            self,
            quant_input,
            quant_hidden_state,
            quant_cell_state,
            quant_weight_ii,
            quant_weight_if,
            quant_weight_ic,
            quant_weight_io,
            quant_weight_hi,
            quant_weight_hf,
            quant_weight_hc,
            quant_weight_ho,
            quant_bias_input,
            quant_bias_forget,
            quant_bias_cell,
            quant_bias_output):
        if self.quantized_cell:
            return self.quantized_cell_symbolic_execution(
                quant_input,
                quant_hidden_state,
                quant_cell_state,
                quant_weight_ii,
                quant_weight_if,
                quant_weight_ic,
                quant_weight_io,
                quant_weight_hi,
                quant_weight_hf,
                quant_weight_hc,
                quant_weight_ho,
                quant_bias_input,
                quant_bias_forget,
                quant_bias_cell,
                quant_bias_output)
        else:
            if onnx_export_opset() < 14:
                raise RuntimeError("Export of float LSTM cell requires at least opset_version=14.")
            # The ONNX standard requires parameters to have shape
            # weight_i: [num_directions, 4*hidden_size, input_size]
            # weight_h: [num_directions, 4*hidden_size, hidden_size]
            # bias: [num_directions, 8*hidden_size]
            quant_weight_i = torch.cat(
                [quant_weight_ii, quant_weight_io, quant_weight_if, quant_weight_ic], dim=0)
            quant_weight_h = torch.cat(
                [quant_weight_hi, quant_weight_ho, quant_weight_hf, quant_weight_hc], dim=0)
            quant_bias = torch.cat(
                [quant_bias_input, quant_bias_output, quant_bias_forget, quant_bias_cell], dim=0)
            # The ONNX standards assumes separate hidden biases, we set them to 0
            quant_bias = torch.cat([quant_bias, torch.zeros_like(quant_bias)], dim=0)
            # Add a leading dimension for the direction
            quant_weight_i = quant_weight_i.unsqueeze(0)
            quant_weight_h = quant_weight_h.unsqueeze(0)
            quant_bias = quant_bias.unsqueeze(0)
            # Compute relevant dimensions
            seq_len = quant_input.size(int(self.symbolic_kwargs['batch_first']))
            hidden_size = int(quant_hidden_state.size(1))
            # sSquence_lens is a tensor of dimension batch size with the sequence length
            # of each element in the batch. We don't support variable sequence length yet
            # so they are set all to the same value
            sequence_lens = torch.empty(
                quant_hidden_state.size(0), dtype=torch.int32).fill_(seq_len)
            # Initial hidden and cell state have an extra direction dimension and
            #  different shapes depending on whether batch_first is set or not
            quant_hidden_state = quant_hidden_state.unsqueeze(0)
            quant_cell_state = quant_cell_state.unsqueeze(0)
            if self.symbolic_kwargs['batch_first']:
                quant_hidden_state = quant_hidden_state.permute(1, 0, 2)
                quant_cell_state = quant_cell_state.permute(1, 0, 2)
            # The ONNX standards requires an extra dimension for the direction,
            # which we don't want here, so we squeeze it. We compute the Squeeze op
            # manually within the function since tracing tend to overcomplicate things
            if self.symbolic_kwargs['batch_first']:
                output_dir_axes = torch.tensor([1], dtype=torch.int64)
                state_dir_axes = torch.tensor([1], dtype=torch.int64)
            else:
                output_dir_axes = torch.tensor([2], dtype=torch.int64)
                state_dir_axes = torch.tensor([0], dtype=torch.int64)
            return LSTMCellFn.apply(
                quant_input,
                quant_weight_i,
                quant_weight_h,
                quant_bias,
                sequence_lens,
                quant_hidden_state,
                quant_cell_state,
                'reverse' if self.symbolic_kwargs['reverse_input'] else 'forward',
                hidden_size,
                self.symbolic_kwargs['cifg'],
                self.symbolic_kwargs['batch_first'],
                output_dir_axes,
                state_dir_axes)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from dataclasses import field
from functools import partial
import operator
from typing import Callable, Dict, List, Optional, Set, Tuple, Union
import warnings

import torch
from torch.fx import GraphModule as TorchGraphModule
import torch.nn as nn

from brevitas.fx import GraphModule
from brevitas.fx import Node
from brevitas.graph.base import GraphTransform
from brevitas.graph.base import ModuleInstanceToModuleInstance
from brevitas.graph.utils import get_module
from brevitas.graph.utils import get_node
from brevitas.nn.equalized_layer import EqualizedModule
from brevitas.nn.equalized_layer import INPUT_NAMES
from brevitas.nn.quant_scale_bias import ScaleBias
from brevitas.utils.torch_utils import KwargsForwardHook

from .base import GraphTransform
from .base import InsertModuleCallAfter

__all__ = ['GraphActivationEqualization', 'LayerwiseActivationEqualization', 'EqualizeGraph']

EPSILON = 1e-9

_supported_layers = (
    nn.ConvTranspose1d,
    nn.ConvTranspose2d,
    nn.ConvTranspose3d,
    nn.MultiheadAttention,
    nn.Conv1d,
    nn.Conv2d,
    nn.Conv3d,
    nn.Linear,
    nn.LayerNorm,
    nn.BatchNorm1d,
    nn.BatchNorm2d,
    nn.BatchNorm3d)

_scale_invariant_layers = (
    nn.Dropout,
    nn.Dropout2d,
    nn.Dropout3d,
    nn.MaxPool1d,
    nn.MaxPool2d,
    nn.MaxPool3d,
    nn.AvgPool1d,
    nn.AvgPool2d,
    nn.AvgPool3d,
    nn.AdaptiveAvgPool1d,
    nn.AdaptiveAvgPool2d,
    nn.AdaptiveAvgPool3d,
    nn.Identity,
    nn.ReLU,
    nn.LeakyReLU)

_scale_invariant_op = (
    torch.mul,
    operator.mul,
    operator.imul,
    operator.__mul__,
    operator.__imul__,
    torch.nn.functional.interpolate)

_select_op = (operator.getitem, operator.__getitem__)

_reshaping_op = ('view', 'reshape', 'flatten', 'contiguous', torch.reshape, torch.flatten)

_scale_varying_activations = (
    torch.nn.Sigmoid, torch.nn.Tanh, torch.nn.ReLU6, torch.nn.GELU, torch.nn.SiLU)

_residual_methods = ('add', 'add_')

_residual_fns = (torch.add, operator.add, operator.iadd, operator.__add__, operator.__iadd__)

_batch_norm = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)

_ignore_ops = (getattr, 'size')


# Start and End identify the starting and ending channels of the weight matrix that need to be
# equalized.
# Offset refers to the relative position of these channels with respect to
# the other matrices' channels that are equalized simultaneously.
# Source matrix are always fully equalized, while sinks can be partially equalized.
@dataclass
class EqualizationIndexes:
    start: int = 0
    end: int = 0
    offset: int = 0


# Required for being hashable
@dataclass(eq=True, frozen=True)
class WeightBiasWrapper:
    weight: torch.Tensor = None
    bias: torch.Tensor = None


# Required for being hashable
@dataclass(eq=True, frozen=True)
class Region:
    srcs: Dict = field(default_factory=dict)
    sinks: Dict = field(default_factory=dict)
    acts: Tuple = field(default_factory=tuple)
    name_to_module: Dict = field(default_factory=dict)

    @property
    def srcs_names(self):
        return [name.split("$")[0] for name in self.srcs.keys()]

    @property
    def sinks_names(self):
        return [name.split("$")[0] for name in self.sinks.keys()]

    def get_module_from_name(self, name: str) -> nn.Module:
        name = name.split("$")[0]
        return self.name_to_module[name]


@dataclass
class WalkRegionState:
    srcs: Dict = field(default_factory=dict)
    sinks: Dict = field(default_factory=dict)
    acts: Set = field(default_factory=set)
    history: set = field(default_factory=set)
    name_to_module: Dict = field(default_factory=dict)

    cat_encoutered: bool = False
    offset: int = 0
    update_offset: bool = False

    @property
    def srcs_names(self):
        return [name.split("$")[0] for name in self.srcs.keys()]

    @property
    def sinks_names(self):
        return [name.split("$")[0] for name in self.sinks.keys() if name is not _UNSUPPORTED_OP]

    def add(
            self,
            type: str,
            name: str,
            module: nn.Module,
            indexes: Optional[EqualizationIndexes] = None):
        if type == 'srcs' or type == 'sinks':
            assert indexes is not None
            full_source_name = name + '$' + str(indexes)
            getattr(self, type)[full_source_name] = indexes
        elif type == 'acts':
            self.acts.add(name)
        self.name_to_module[name] = module

    def add_srcs(self, src_name: str, src: nn.Module, indexes: EqualizationIndexes):
        self.add('srcs', src_name, src, indexes)

    def add_sinks(self, sink_name: str, sink: nn.Module, indexes: EqualizationIndexes):
        self.add('sinks', sink_name, sink, indexes)

    def add_acts(self, act_name: str, act: nn.Module):
        self.add('acts', act_name, act)

    def get_module_from_name(self, name: str) -> nn.Module:
        name = name.split("$")[0]
        return self.name_to_module[name]


def __str__(self):
    return str(self.start) + '_' + str(self.end) + '_' + str(self.offset)


_UNSUPPORTED_OP = object()


def _select_scale_computation_fn(
        scale_computation_type: str) -> Callable[[torch.Tensor], torch.Tensor]:
    if scale_computation_type == 'maxabs':
        return _channel_maxabs
    elif scale_computation_type == 'range':
        return _channel_range
    else:
        raise RuntimeError(f"Scale computation type {scale_computation_type} not supported")


class activation_equalization_mode:

    def __init__(
            self,
            model,
            alpha,
            add_mul_node=True,
            layerwise=True,
            enabled=True,
            blacklist_layers=None,
            co_optimize_act_weights=False) -> None:
        self.model = model
        self.alpha = alpha
        self.enabled = enabled
        self.add_mul_node = add_mul_node
        self.co_optimize_act_weights = co_optimize_act_weights
        if layerwise:
            if not self.add_mul_node:
                raise ValueError("Layerwise activation equalization requires add_mul_node")
            self.graph_act_eq = LayerwiseActivationEqualization(
                self.model, blacklist_layers=blacklist_layers)
        else:
            if not isinstance(self.model, (TorchGraphModule, GraphModule)):
                raise TypeError(
                    "A Torch FX representation of the model is needed for Graph Activation Equalization"
                )
            self.graph_act_eq = GraphActivationEqualization(
                self.model, self.add_mul_node, co_optimize_act_weights=co_optimize_act_weights)
        self.scale_factors = None

    def __enter__(self):
        if self.enabled:
            self.graph_act_eq.setup()
        return self

    def __exit__(self, type, value, traceback):
        if self.enabled:
            self.scale_factors = self.graph_act_eq.apply(self.alpha)
        return True  # To propagate exceptions


def _channel_range(inp: torch.Tensor, dim: int = 1) -> torch.Tensor:
    mins, _ = inp.min(dim=dim)
    maxs, _ = inp.max(dim=dim)
    out = maxs - mins
    # correct corner case where where all weights along a channel have the same value
    # e.g. when a mean/nn.AvgPool/nn.AdaptiveAvgPool is converted to a depth-wise conv
    out = torch.where(out == 0., torch.mean(inp, dim=dim), out)

    # convert to positive range, in case any of the values are negative,
    # highly likely in cases when there is only one value per channel, such as in Batch Norm
    out = torch.abs(out)
    return out


def _channel_maxabs(inp: torch.Tensor, dim: int = 1) -> torch.Tensor:
    out = torch.max(torch.abs(inp), dim=dim)[0]
    return out


def _get_input_axis(module: nn.Module) -> Optional[int]:
    """
    Given a sink module, determine the axis associated to the input channels.
    Return None if not supported.
    """
    if isinstance(module, (nn.Linear, nn.MultiheadAttention)):
        return 1
    elif isinstance(module, _batch_norm):
        return 0
    elif isinstance(module, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
        if module.groups == 1:
            return 1
        elif module.groups == module.out_channels:
            return 0
    elif isinstance(module, (nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d)):
        if module.groups == 1:
            return 0
        elif module.groups == module.out_channels:
            return 1
    elif isinstance(module, nn.LayerNorm):
        # We assume normalization happens only along the channel dimension
        if len(module.weight.shape) == 1:
            return 0
        else:
            return None
    else:
        return None


def _get_output_axis(module: nn.Module) -> Optional[int]:
    """
    Given a source module, determine the axis associated to the output channels.
    Return None if not supported.
    """
    if isinstance(module,
                  (nn.Linear,
                   nn.Conv1d,
                   nn.Conv2d,
                   nn.Conv3d,
                   nn.MultiheadAttention,
                   nn.BatchNorm1d,
                   nn.BatchNorm2d,
                   nn.BatchNorm3d)):
        return 0
    elif isinstance(module, (nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d)):
        return 1
    elif isinstance(module, nn.LayerNorm):
        # We assume normalization happens only along the channel dimension
        if len(module.weight.shape) == 1:
            return 0
        else:
            return None
    else:
        return None


def _get_act_axis(sink_module: nn.Module) -> Optional[int]:
    """
    Given a layer, return activation axis associated to the feature dim.
    Even though BatchNorm layers are supported, it gives no information about the feature dim.
    """
    if isinstance(sink_module, (nn.Linear, nn.MultiheadAttention)):
        return -1
    elif isinstance(sink_module,
                    (nn.Conv1d,
                     nn.Conv2d,
                     nn.Conv3d,
                     nn.ConvTranspose1d,
                     nn.ConvTranspose2d,
                     nn.ConvTranspose3d)):
        return 0
    else:
        return None


def _combine_weights_bias(
        weight: torch.Tensor, bias_shrinkage: Union[float, str],
        bias: Optional[torch.Tensor]) -> torch.Tensor:
    """Combine weights and bias before graph equalizattion
    This method merges the weight and bias of the sources, so that the resulting equalizer scale factor
    is influenced also by the magnitude of the bias, mitigated by a shrink factor.
    This technique avoids that, after the equalization procedure, the bias values become too big,
    negatively impacting the quantization accuracy.
    The bias shrinkage factor regulates how much the bias magnitude will affect the subsequence
    calculation of the scale.
    """
    if bias is None:
        return weight.data

    bias = bias.data
    weight = weight.reshape(weight.shape[0], -1)
    bias = bias.reshape(-1, 1)

    weight = torch.where(
        torch.abs(weight) <= EPSILON, torch.tensor(EPSILON).type_as(weight), weight)
    factor = torch.abs(bias) / torch.abs(weight)

    # From https://github.com/Xilinx/Vitis-AI/blob/master/src/vai_quantizer/vai_q_pytorch/nndct_shared/optimization/commander.py#L450
    if bias_shrinkage == 'vaiq':
        if torch.abs(bias).max() < 10 and (torch.abs(bias).max() / torch.abs(weight).max()) < 20:
            if torch.median(factor) > 100 or torch.mean(factor) > 1000:
                shrink_factor = 5.
            else:
                shrink_factor = 2.
        else:
            if torch.median(factor) > 30 or torch.mean(factor) > 500:
                shrink_factor = 20.
            elif torch.median(factor) > 15 or torch.mean(factor) > 250:
                shrink_factor = 10.
            else:
                shrink_factor = 5.
    elif isinstance(bias_shrinkage, (int, float)):
        shrink_factor = bias_shrinkage
    else:
        raise RuntimeError(f"{bias_shrinkage} not supported.")
    weight_bias = torch.cat([weight, bias / shrink_factor], 1)
    return weight_bias


def transpose(tensor: torch.Tensor, axis: int):
    """
    Given a tensor and an axis, this function re-arranges the tensor so that the axis and
    the first dimension are swapped.
    """
    shape = list(range(tensor.ndim))
    axis = shape[axis]
    shape.insert(0, axis)
    del shape[axis + 1]
    return tensor.permute(shape)


def _cross_layer_equalization(
        region: Region,
        merge_bias: bool,
        scale_computation_type: str,
        bias_shrinkage: Optional[Union[float, str]] = None,
        list_of_act_val: Optional[torch.Tensor] = None,
        list_of_insert_mul_node_fn: Optional[List[Callable]] = None,
        alpha: float = 0.5,
        co_optimize_act_weights: bool = False) -> torch.Tensor:
    """
    Given two adjacent tensors', the weights are scaled such that
    the ranges of the first tensors' output channel are equal to the
    ranges of the second tensors' input channel
    """

    # If equalization criteria are not met, we return a scalar one to indicate that no equalization
    # has been performed
    def _no_equalize():
        return torch.tensor(1., dtype=dtype)

    # If a module has `allocate_params` attribute, we must load the weights following that method

    for name in (region.srcs_names + region.sinks_names):
        module = region.get_module_from_name(name)
        if hasattr(module, 'allocate_params'):
            module.allocate_params(module)

    act_sink_axes = {}
    act_sources_axes = {}
    single_module = region.get_module_from_name(next(iter(region.sinks_names)))
    dtype = next(single_module.parameters()).dtype

    max_shape_srcs = 0
    for name, indexes in region.srcs.items():
        max_shape_srcs = max(max_shape_srcs, indexes.end + indexes.offset)
    max_shape_sinks = 0
    for name, indexes in region.sinks.items():
        max_shape_sinks = max(max_shape_sinks, indexes.offset + (indexes.end - indexes.start))

    # Exit if source and sink have different sizes
    if max_shape_srcs != max_shape_sinks and len(region.srcs) > 0:
        return _no_equalize()

    src_axes = {}
    for name, indexes in region.srcs.items():
        module = region.get_module_from_name(name)
        # If module is not supported, do not perform graph equalization
        axis = _get_output_axis(module)
        act_sources_axes[name] = _get_act_axis(module)
        if not isinstance(module, _supported_layers):
            return _no_equalize()
        if isinstance(module, nn.MultiheadAttention):
            module = module.out_proj
        src_axes[name] = (module, axis)

    sink_axes = {}
    for name, indexes in region.sinks.items():
        module = region.get_module_from_name(name)
        axis = _get_input_axis(module)
        act_sink_axes[name] = _get_act_axis(module)
        # If module is not supported, do not perform graph equalization
        if not isinstance(module, _supported_layers) or module in _batch_norm:
            return _no_equalize()
        # For MultiheadAttention, we support only self-attetion
        if isinstance(module, nn.MultiheadAttention) and module.in_proj_weight is not None:
            # For sinks, we only need to modify the weight but not the bias
            module = WeightBiasWrapper(module.in_proj_weight)
        elif isinstance(module, nn.MultiheadAttention) and module.in_proj_weight is None:
            return _no_equalize()
        sink_axes[name] = (module, axis)
    # If act_val is enabled, use source or sink weights to determine the activation channel
    # For example, if the source is BatchNorm, we need to use the information coming from the sinks
    if list_of_act_val is not None:
        list_of_sink_axes = [x for x in list(act_sink_axes.values()) if x is not None]
        list_of_source_axes = [x for x in list(act_sources_axes.values()) if x is not None]
        if len(list_of_sink_axes) > 0:
            act_axis = list_of_sink_axes[0]
        elif len(list_of_source_axes) > 0:
            act_axis = list_of_source_axes[0]
        else:
            return _no_equalize()
        # If there is a mismatch in the activation channel (e.g. a transpose/flatten op in between),
        # do not perform equalization
        if any([act_axis != axis for axis in list_of_source_axes + list_of_sink_axes]):
            return _no_equalize()

    # Check if any of the axis is None, which means that the module is not supported.
    # In that case, do not perform graph equalization
    axes_to_check = [axis for _, axis in list(src_axes.values()) + list(sink_axes.values())]
    if None in axes_to_check:
        return _no_equalize()

    scale_fn = _select_scale_computation_fn(scale_computation_type)
    sink_weights = {
        name: transpose(m.weight.cpu().to(torch.float32), axis)
        for name, (m, axis) in sink_axes.items()}
    srcs_range = -1 * torch.ones(max_shape_srcs, device='cpu', dtype=torch.float32)
    sinks_range = -1 * torch.ones(max_shape_sinks, device='cpu', dtype=torch.float32)
    for k, v in sink_weights.items():
        # Sinks can be partially equalized, thus we need to select
        # only the channels we are interested in
        indexes = region.sinks[k]
        # Compute the range of the channels we need to equalize
        weight_range = scale_fn(v.reshape(v.size(0), -1))[indexes.start:indexes.end]
        # Compute the numbers of channels we are equalizing
        channel_range = indexes.end - indexes.start
        # Use the offset and the range to update the correct range in the sinks
        sinks_range[indexes.offset:indexes.offset + channel_range] = torch.max(
            sinks_range[indexes.offset:indexes.offset + channel_range], weight_range)

    # Determine the srcs_range based on where we are performing activation equalization or
    # weight equalization
    if merge_bias:
        src_weights = {
            name: _combine_weights_bias(transpose(m.weight, axis), bias_shrinkage,
                                        m.bias).cpu().to(torch.float32)
            for name, (m, axis) in src_axes.items()}
    else:
        src_weights = {
            name: transpose(m.weight.cpu().to(torch.float32), axis)
            for name, (m, axis) in src_axes.items()}
    for k, v in src_weights.items():
        # Srcs are always fully equalized, thus we simply need to apply the offset to position them
        # correctly with respect to the other srcs matrices.
        indexes = region.srcs[k]
        channel_start = indexes.offset + indexes.start
        channel_end = indexes.offset + indexes.end
        weight_range = scale_fn(v.reshape(v.size(0), -1))
        srcs_range[channel_start:channel_end] = torch.max(
            srcs_range[channel_start:channel_end], weight_range)
    if list_of_act_val is not None:
        list_of_act_val_shapes = [act_val.shape for act_val in list_of_act_val]
        if len(list_of_act_val_shapes) > 0:
            shape_0 = list_of_act_val_shapes[0]
            if any(shape_0 != shape for shape in list_of_act_val_shapes):
                return _no_equalize()
        list_of_act_val = list_of_act_val = [
            transpose(act_val, act_axis) for act_val in list_of_act_val]
        srcs_range_act = scale_fn(
            torch.cat([
                act_val.reshape(act_val.size(0), -1).cpu().to(torch.float32)
                for act_val in list_of_act_val],
                      1))

    if list_of_act_val is not None:
        if co_optimize_act_weights and len(src_axes) > 0:
            srcs_range = .5 * srcs_range + .5 * srcs_range_act
        else:
            srcs_range = srcs_range_act

    # If there is a mismatch between srcs and sinks values, exit
    if srcs_range.shape != sinks_range.shape:
        warnings.warn(
            "Detected source and sink with non compatible shapes, equalization is skipped")
        return _no_equalize()

    # Instead of clipping very low values, which would cause their reciprocal to be very large
    # thus hindering quantization, we set both sources and sinks to one,
    # which is the no-op equivalent for equalization.
    channelwise_no_equalize = (sinks_range <= EPSILON) | (srcs_range <= EPSILON)
    sinks_range = torch.where(
        channelwise_no_equalize, torch.tensor(1., dtype=torch.float32, device='cpu'), sinks_range)
    srcs_range = torch.where(
        channelwise_no_equalize, torch.tensor(1., dtype=torch.float32, device='cpu'), srcs_range)

    srcs_range = torch.pow(srcs_range, alpha)
    sinks_range = torch.pow(sinks_range, 1 - alpha)
    scaling_factors = srcs_range / sinks_range
    inverse_scaling_factors = torch.reciprocal(scaling_factors)

    if list_of_act_val is not None and list_of_insert_mul_node_fn is not None:
        device = list_of_act_val[0].device
        for act_val_shape, insert_mul_node_fn in zip(list_of_act_val_shapes, list_of_insert_mul_node_fn):
            insert_mul_node_fn(
                inverse_scaling_factors.to(device=device, dtype=dtype), act_val_shape, act_axis)
    if len(src_axes) > 0:
        for name, (module, axis) in src_axes.items():
            module_device = module.weight.device
            indexes = region.srcs[name]
            channel_start = indexes.offset + indexes.start
            channel_end = indexes.offset + indexes.end
            partial_inverse_scale = inverse_scaling_factors[channel_start:channel_end].to(
                device=module_device, dtype=dtype)
            if hasattr(module, 'bias') and module.bias is not None:
                _update_weights(
                    module, module.bias * partial_inverse_scale.view_as(module.bias), attr='bias')
            src_broadcast_size = [1] * module.weight.ndim
            src_broadcast_size[axis] = module.weight.size(axis)

            _update_weights(
                module,
                module.weight * torch.reshape(partial_inverse_scale, src_broadcast_size),
                attr='weight')
    for name, (module, axis) in sink_axes.items():
        module_device = module.weight.device
        sink_broadcast_size = [1] * module.weight.ndim
        sink_broadcast_size[axis] = module.weight.size(axis)
        indexes = region.sinks[name]
        channel_range = indexes.end - indexes.start
        partial_scaling = torch.ones(module.weight.size(axis), device='cpu', dtype=dtype)
        # We replace the scaling factors of the channels we need to equalize, leaving the other to
        # one (i.e., no equalization)
        partial_scaling[indexes.start:indexes.end] = scaling_factors[indexes.offset:indexes.offset +
                                                                     channel_range]
        partial_scaling = partial_scaling.to(device=module_device, dtype=dtype)
        _update_weights(
            module,
            module.weight * torch.reshape(partial_scaling, sink_broadcast_size),
            attr='weight')

    # If a module has `offload_params` attribute, we must offload the weights following that method
    for name in (region.srcs_names + region.sinks_names):
        module = region.get_module_from_name(name)
        if hasattr(module, 'offload_params'):
            module.offload_params(module)

    return scaling_factors


def _update_weights(original_module, new_value, attr='weight'):
    if isinstance(original_module, WeightBiasWrapper):
        setattr(getattr(original_module, attr), 'data', new_value)
    else:
        setattr(original_module, attr, nn.Parameter(new_value))


def _equalize(
        model: GraphModule,
        regions: Set[Tuple[str]],
        iterations: int,
        threshold: float,
        merge_bias: bool,
        bias_shrinkage: Union[str, float],
        scale_computation_type: str) -> GraphModule:
    """
    Generalized version of section 4.1 of https://arxiv.org/pdf/1906.04721.pdf
    """
    for i in range(iterations):
        scale_factor_max = None
        for region in regions:
            scale_factors_region = _cross_layer_equalization(
                region,
                merge_bias=merge_bias,
                bias_shrinkage=bias_shrinkage,
                scale_computation_type=scale_computation_type)
            scale_factor_region_max = torch.max(torch.abs(1 - scale_factors_region))
            if scale_factor_max is not None:
                scale_factor_max = torch.max(scale_factor_max, scale_factor_region_max)
            else:
                scale_factor_max = scale_factor_region_max
        if threshold is not None and scale_factor_max < threshold:
            break
    return model


def _is_supported_module(graph_model: GraphModule, node: Node) -> bool:
    if node.op == 'call_module':
        module = get_module(graph_model, node.target)
        if isinstance(module, _supported_layers):
            # We support only self-attention
            if isinstance(module, nn.MultiheadAttention):
                kwargs = dict(node.kwargs)
                # When using hf/accelerate, we need to check the signature of the original forward
                forward_to_check = module._old_forward if hasattr(
                    module, '_old_forward') else module.forward
                kwargs.update(zip(forward_to_check.__code__.co_varnames[1:], node.args))
                return kwargs['query'].name == kwargs['key'].name == kwargs['value'].name
            return True
    return False


def _is_scale_invariant_module(graph_model: GraphModule, node: Node) -> bool:
    return node.op == 'call_module' and isinstance(
        get_module(graph_model, node.target), _scale_invariant_layers)


def _is_scale_varying_activation(graph_model, node):
    return node.op == 'call_module' and isinstance(
        get_module(graph_model, node.target), _scale_varying_activations)


def _is_scale_invariant_function(node: Node) -> bool:
    out = node.op == 'call_function' and node.target in _scale_invariant_op + _select_op
    if node.target == torch.nn.functional.interpolate:
        out &= node.kwargs.get('mode', None) == 'nearest'
    return out


def _is_reshaping_op(node: Node) -> bool:
    return node.target in _reshaping_op


def get_weight_source(module):
    transpose = lambda weight, axis: weight if axis == 0 else weight.transpose(0, 1)
    if isinstance(module, nn.MultiheadAttention) and not hasattr(module, 'out_proj'):
        raise RuntimeError("Configuration for Multiheadattention not supported")
    weight = module.out_proj.weight if isinstance(module, nn.MultiheadAttention) else module.weight
    axis = _get_output_axis(module)
    weight = transpose(weight, axis)
    return weight


def get_weight_sink(module):
    transpose = lambda weight, axis: weight if axis == 0 else weight.transpose(0, 1)
    if isinstance(module, nn.MultiheadAttention) and not hasattr(module, 'in_proj_weight'):
        raise RuntimeError("Configuration for Multiheadattention not supported")
    weight = WeightBiasWrapper(module.in_proj_weight).weight if isinstance(
        module, nn.MultiheadAttention) else module.weight
    axis = _get_input_axis(module)
    weight = transpose(weight, axis)
    return weight


def find_srcs_channel_dim(model, inp_node):
    if _is_supported_module(model, inp_node):
        # If we meet a supported module, determine the channel shape
        module = get_module(model, inp_node.target)
        # Since we are walking up, we consider the module as srcs
        weight = get_weight_source(module)
        channel = weight.shape[0]
        return channel
    elif _is_add(inp_node):
        all_channels = []
        for n in inp_node.all_input_nodes:
            all_channels.append(find_srcs_channel_dim(model, n))
        # All branches to add should have the same amount of channels
        if all([channel == all_channels[0] for channel in all_channels]):
            return all_channels[0]
        else:
            return _UNSUPPORTED_OP
    elif _is_cat(inp_node):
        total_channels = 0
        # If it's cat, we need to sum the channel shape of all the branches
        for n in inp_node.all_input_nodes:
            total_channels += find_srcs_channel_dim(model, n)
        return total_channels
    elif _is_scale_invariant_module(model, inp_node) or _is_scale_invariant_function(inp_node):
        return find_srcs_channel_dim(model, inp_node.all_input_nodes[0])
    else:
        return _UNSUPPORTED_OP


def cat_handler(graph_model: GraphModule, starting_node: Node, state: WalkRegionState):

    state.srcs.clear()
    state.sinks.clear()
    state.history.clear()
    # Keep track that concatenation has been encoutered once
    state.cat_encoutered = True
    state.update_offset = True
    state.offset = 0
    find_srcs(graph_model, starting_node, state)
    state.update_offset = False
    state.offset = 0
    find_sinks(graph_model, starting_node, state)


def _is_cat(node):
    return node.target in (torch.cat,)


def _is_add(node):
    return (
        node.op == 'call_method' and node.target in _residual_methods or
        node.op == 'call_function' and node.target in _residual_fns)


def find_srcs(graph_model: GraphModule, starting_node: Node,
              state: WalkRegionState) -> Dict[str, Set]:
    node_list = starting_node.all_input_nodes
    update_offset_state = state.update_offset
    for node in node_list:
        # we keep a history of how the graph has been walked already, invariant to the direction,
        # to avoid getting stuck in a loop
        path = (node, starting_node)
        if path not in state.history:
            state.history.add(path)
        else:
            continue
        if _is_supported_module(graph_model, node):
            module = get_module(graph_model, node.target)
            weight = get_weight_source(module)
            eq_indexes = EqualizationIndexes(0, weight.shape[0], state.offset)

            # After we found a source, we need to check if it branches into multiple sinks
            state.add_srcs(node.target, module, eq_indexes)
            find_sinks(graph_model, node, state)
            state.offset = state.offset if not state.update_offset else state.offset + weight.shape[
                0]
        elif _is_scale_invariant_module(
                graph_model, node) or _is_scale_invariant_function(node) or _is_reshaping_op(node):
            find_sinks(graph_model, node, state)
            find_srcs(graph_model, node, state)
        elif (node.op == 'call_method' and node.target in _residual_methods or
              node.op == 'call_function' and node.target in _residual_fns):
            state.update_offset = False
            find_sinks(graph_model, node, state)
            find_srcs(graph_model, node, state)
            state.update_offset = update_offset_state
        elif _is_cat(node):
            # The first time we encoutered a cat differes from all subsequent ones
            if not state.cat_encoutered:
                # We restart the region search starting from the cat
                cat_handler(graph_model, node, state)
            else:
                state.update_offset = False
                find_sinks(graph_model, node, state)
                state.update_offset = True
                find_srcs(graph_model, node, state)
                state.update_offset = update_offset_state
        elif node.target in _ignore_ops:
            continue
        else:
            # If we meet an unrecognized op, we add None to invalidate the region
            state.sinks[_UNSUPPORTED_OP] = _UNSUPPORTED_OP


def find_sinks(graph_model: GraphModule, starting_node: Node,
               state: WalkRegionState) -> Dict[str, Set]:
    node_list = starting_node.users
    update_offset_state = state.update_offset
    for node in node_list:
        # we keep a history of how the graph has been walked already, invariant to the direction,
        # to avoid getting stuck in a loop
        # Note that the path is inverted with respect to find_srcs
        path = (starting_node, node)
        if path not in state.history:
            state.history.add(path)
        else:
            continue
        if _is_supported_module(graph_model, node):
            module = get_module(graph_model, node.target)
            weight = get_weight_sink(module)
            eq_indexes = EqualizationIndexes(0, weight.shape[0], state.offset)
            # It is not possible to equalize through LayerNorm as sink
            if isinstance(module, (nn.LayerNorm,) + _batch_norm):
                state.sinks[_UNSUPPORTED_OP] = _UNSUPPORTED_OP
            else:
                state.add_sinks(node.target, module, eq_indexes)

        elif _is_scale_invariant_module(
                graph_model, node) or _is_scale_invariant_function(node) or _is_reshaping_op(node):
            find_sinks(graph_model, node, state)
        elif (node.op == 'call_method' and node.target in _residual_methods or
              node.op == 'call_function' and node.target in _residual_fns):
            state.update_offset = False
            find_sinks(graph_model, node, state)
            find_srcs(graph_model, node, state)
            state.update_offset = update_offset_state
        elif _is_cat(node):
            # The first time we encoutered a cat differes from all subsequent ones
            if not state.cat_encoutered:
                # We restart the region search starting from the cat
                cat_handler(graph_model, node, state)
            else:
                # In this case we define all our sinks, and isolate only the channels we want
                # to equalize (start, end).
                # Furthermore, we need to consider the offset given by the sources of the second cat
                index = node.all_input_nodes.index(starting_node)
                channels = []
                for n in node.all_input_nodes:
                    channel_dim = find_srcs_channel_dim(graph_model, n)
                    channels.append(channel_dim)

                # If we found an unsupported op while walking up, we exit this branch and
                # invalidate the region
                if _UNSUPPORTED_OP in channels:
                    state.sinks[_UNSUPPORTED_OP] = _UNSUPPORTED_OP
                    continue
                start = sum(channels[:index])
                end = start + channels[index]
                new_state = WalkRegionState(offset=state.offset)
                find_sinks(graph_model, node, new_state)

                for k in new_state.sinks_names:
                    state.add_sinks(
                        k,
                        new_state.get_module_from_name(k),
                        EqualizationIndexes(start, end, new_state.offset))
                state.srcs.update(new_state.srcs)
        elif node.target in _ignore_ops:
            continue
        else:
            # If we meet an unrecognized op, we add None to invalidate the region
            state.sinks[_UNSUPPORTED_OP] = _UNSUPPORTED_OP


def _extract_regions(
        graph_model: GraphModule,
        add_mul_node: bool = False,
        return_acts: bool = False) -> List[Region]:
    regions = list()
    for node in graph_model.graph.nodes:
        if _is_supported_module(graph_model,
                                node) or (add_mul_node and
                                          _is_scale_varying_activation(graph_model, node)):
            state = WalkRegionState()
            if _is_scale_varying_activation(graph_model, node):
                module = get_module(graph_model, node.target)
                state.add_acts(node.target, module)
            else:
                module = get_module(graph_model, node.target)
                weight = get_weight_source(module)
                eq_indexes = EqualizationIndexes(0, weight.shape[0], 0)
                state.add_srcs(node.target, module, eq_indexes)
            find_sinks(graph_model, node, state)
            if len(state.sinks) > 0 and _UNSUPPORTED_OP not in state.sinks.keys():
                sorted_srcs = dict(sorted(state.srcs.items()))
                sorted_sinks = dict(sorted(state.sinks.items()))
                sorted_acts = tuple(sorted(state.acts))
                if return_acts:
                    region = Region(
                        srcs=sorted_srcs,
                        sinks=sorted_sinks,
                        acts=sorted_acts,
                        name_to_module=state.name_to_module)
                else:
                    region = Region(
                        srcs=sorted_srcs, sinks=sorted_sinks, name_to_module=state.name_to_module)

                if region not in regions:
                    regions.append(region)
    return regions


class EqualizeGraph(GraphTransform):

    def __init__(
            self,
            iterations: int = 10,
            threshold: float = 0.05,
            return_regions: bool = False,
            merge_bias: bool = True,
            bias_shrinkage: Union[float, str] = 'vaiq',
            scale_computation_type: str = 'maxabs') -> None:
        super(EqualizeGraph, self).__init__()
        self.iterations = iterations
        self.return_regions = return_regions
        self.merge_bias = merge_bias
        self.bias_shrinkage = bias_shrinkage
        self.threshold = threshold
        self.scale_computation_type = scale_computation_type

    def apply(self,
              graph_model: GraphModule) -> Union[Tuple[GraphModule, Set[Tuple[str]]], GraphModule]:
        regions = _extract_regions(graph_model)
        if len(regions) > 0:
            graph_model = _equalize(
                graph_model,
                regions,
                self.iterations,
                self.threshold,
                self.merge_bias,
                self.bias_shrinkage,
                self.scale_computation_type)
        if self.return_regions:
            return graph_model, regions
        else:
            return graph_model


class ActivationEqualization(GraphTransform, ABC):

    def __init__(
            self, model: Union[nn.Module, GraphModule], scale_computation_type: str = 'maxabs'):
        self.model = model
        self.scale_computation_type = scale_computation_type

    @abstractmethod
    def setup(self):
        pass

    @abstractmethod
    def insert_mul_node(self):
        pass

    def create_mul_node(self, scale, shape, axis, batch_dim=0):
        broadcastable_shape = [1] * len(shape)
        broadcastable_shape[axis] = shape[axis]
        # Add Batch Dim
        broadcastable_shape.insert(batch_dim, 1)
        mul_factor = ScaleBias(
            num_features=shape[axis], bias=False, runtime_shape=broadcastable_shape)
        mul_factor.weight.data = scale
        return mul_factor

    def forward_stats_hook(self, module, *args, name, batch_dim=0, use_inp=True, **kwargs):
        # Check for MHA Cross attention, and if found, skip it
        # When using hf/accelerate, we need to check the signature of the original forward
        forward_to_check = module._old_forward if hasattr(
            module, '_old_forward') else module.forward
        kwargs.update(zip(forward_to_check.__code__.co_varnames[1:], args[:-1]))
        if 'query' in kwargs and 'key' in kwargs and 'value' in kwargs:
            if kwargs['query'].data_ptr() != kwargs['key'].data_ptr() != kwargs['value'].data_ptr():
                self.float_act_map[name] = None
                return

        input_kwarg = [x for x in kwargs.keys() if x in INPUT_NAMES][0]
        if use_inp:
            x = kwargs[input_kwarg]
        elif not use_inp:
            x = args[-1]

        # Extra check for batch_dim
        if hasattr(x, 'names') and 'N' in x.names:
            batch_dim = x.names.index('N')

        self.batch_dim_act_map[name] = batch_dim

        dtype = x.dtype
        input_scales = self.scale_fn(x.to(torch.float32), dim=batch_dim).to(dtype)
        if name not in self.float_act_map:
            self.float_act_map[name] = input_scales
        else:
            self.float_act_map[name] = torch.max(self.float_act_map[name], input_scales)

    def remove_hooks(self):
        for hook in self.hooks:
            ModuleInstanceToModuleInstance(hook, hook.module).apply(self.model)


class LayerwiseActivationEqualization(ActivationEqualization):

    def __init__(
            self,
            model,
            scale_computation_type: str = 'maxabs',
            blacklist_layers: Optional[List[str]] = None):
        super(LayerwiseActivationEqualization, self).__init__(model, scale_computation_type)
        self.float_act_map = {}
        self.batch_dim_act_map = {}
        self.hooks = []
        self.add_mul_node = True
        self.blacklist_layers = blacklist_layers

        regions: List[Region] = []
        self.find_module(model, regions)
        self.regions = regions

        if self.scale_computation_type == 'maxabs':
            self.scale_fn = _channel_maxabs
        elif self.scale_computation_type == 'range':
            self.scale_fn = _channel_range

    def find_module(self, model, regions: List, prefix=''):
        """
        Iterate through the model looking at immediate children of every module to look for supported modules.
        This allows us to stop the search when we meet a top-level module that is supported.
        """
        if isinstance(model,
                      _supported_layers) and not isinstance(model, _batch_norm + (nn.LayerNorm,)):
            if self.blacklist_layers is not None and prefix in self.blacklist_layers:
                return
            weight = get_weight_sink(model)
            eq_indexes = EqualizationIndexes(0, weight.shape[0], 0)
            region = Region(sinks={'sinks0': eq_indexes}, name_to_module={'sinks0': model})
            regions.append(region)
        else:
            for name, module in model.named_children():
                full_name = prefix + '.' + name if prefix != '' else name
                self.find_module(module, regions, full_name)

    def setup(self):
        for region in self.regions:
            module = region.get_module_from_name('sinks0')
            batch_dim = 0
            if hasattr(region, 'batch_first'):
                batch_dim = 0 if region.batch_first else 1

            hook_fn = partial(
                self.forward_stats_hook, name=module, batch_dim=batch_dim, use_inp=True)
            new_instance = KwargsForwardHook(module, hook_fn)
            ModuleInstanceToModuleInstance(module, new_instance).apply(self.model)
            self.hooks.append(new_instance)

    def apply(self, alpha):
        scale_factors = []
        self.remove_hooks()
        for region in self.regions:
            module = region.get_module_from_name('sinks0')
            if self.float_act_map[module] == None:
                continue
            insert_mul_fn = partial(
                self.insert_mul_node, region=module, batch_dim=self.batch_dim_act_map[module])
            scale_factors.append(
                _cross_layer_equalization(
                    region,
                    False,
                    scale_computation_type=self.scale_computation_type,
                    list_of_act_val=[self.float_act_map[module]],
                    list_of_insert_mul_node_fn=[insert_mul_fn],
                    alpha=alpha))
        return scale_factors

    def insert_mul_node(self, scale, shape, axis, region, batch_dim=0):
        mul_factor = self.create_mul_node(scale, shape, axis, batch_dim)
        rewriter = ModuleInstanceToModuleInstance(
            region, EqualizedModule(scale_module=mul_factor, layer=region))
        rewriter.apply(self.model)


class GraphActivationEqualization(ActivationEqualization):

    def __init__(
            self,
            model: GraphModule,
            add_mul_node: bool = False,
            scale_computation_type: str = 'maxabs',
            co_optimize_act_weights: bool = False):
        super(GraphActivationEqualization, self).__init__(model, scale_computation_type)
        self.float_act_map = {}
        self.batch_dim_act_map = {}
        self.hooks = []
        self.hooked_modules = set()
        self.add_mul_node = add_mul_node
        self.co_optimize_act_weights = co_optimize_act_weights
        self.regions = _extract_regions(model, add_mul_node=add_mul_node, return_acts=True)

        if self.scale_computation_type == 'maxabs':
            self.scale_fn = _channel_maxabs
        elif self.scale_computation_type == 'range':
            self.scale_fn = _channel_range

    def setup(self):
        # Select only regions with activation to equalize through.
        # If a region has multiple scale varying activation, must also be dropped
        # because we can't propagate scaling factors
        regions_to_drop = []
        for region in self.regions:
            # This condition is for redudancy, since
            # a region with two scale-varying activations cannot be detected in the first place
            if len(region.acts) > 1 and any([isinstance(region.get_module_from_name(act_name),
                                                        _scale_varying_activations)
                                             for act_name in region.acts]):
                regions_to_drop.append(region)
                continue

            # We assume that the entire region has a unique batch_dim
            batch_dim = 0
            for name in region.srcs:
                module = region.get_module_from_name(name)
                if hasattr(module, 'batch_first') and not module.batch_first:
                    batch_dim = 1
            for name in region.sinks:
                module = region.get_module_from_name(name)
                if hasattr(module, 'batch_first') and not module.batch_first:
                    batch_dim = 1

            region_to_search = region.sinks_names if len(region.acts) == 0 else region.acts
            for name in region_to_search:
                module = region.get_module_from_name(name)
                if module not in self.hooked_modules:
                    self.hooked_modules.add(module)
                    use_inp = True if region_to_search == region.sinks_names else False
                    hook_fn = partial(
                        self.forward_stats_hook, name=name, batch_dim=batch_dim, use_inp=use_inp)
                    new_instance = KwargsForwardHook(module, hook_fn)
                    ModuleInstanceToModuleInstance(module, new_instance).apply(self.model)
                    self.hooks.append(new_instance)

        self.regions = [x for x in self.regions if x not in regions_to_drop]

    def apply(self, alpha):
        scale_factors = []
        self.remove_hooks()
        for region in self.regions:
            region_names = region.sinks_names if len(region.acts) == 0 else region.acts
            if any([self.float_act_map[name] is None for name in region_names]):
                continue
            act_module = [region.get_module_from_name(act_name) for act_name in region.acts]
            list_of_act_val = [self.float_act_map[name] for name in region_names]

            list_of_insert_mul_node_fn = None
            if self.add_mul_node and any([
                    isinstance(act, _scale_varying_activations) for act in act_module]):
                # Even though we iterate, this list will always have a single element by definition
                list_of_insert_mul_node_fn = []
                for act_name in region.acts:
                    act_node = get_node(self.model, act_name)
                    list_of_insert_mul_node_fn.append(
                        partial(
                            self.insert_mul_node,
                            act_node=act_node,
                            batch_dim=self.batch_dim_act_map[act_name]))

            scale_factors.append(
                _cross_layer_equalization(
                    region,
                    False,
                    scale_computation_type=self.scale_computation_type,
                    list_of_act_val=list_of_act_val,
                    list_of_insert_mul_node_fn=list_of_insert_mul_node_fn,
                    alpha=alpha,
                    co_optimize_act_weights=self.co_optimize_act_weights))

        return scale_factors

    def insert_mul_node(self, scale, shape, axis, act_node, batch_dim=0):
        mul_factor = self.create_mul_node(scale, shape, axis, batch_dim)
        mul_factor_name = act_node.name + 'act_eq_mul'
        self.model.add_module(mul_factor_name, mul_factor)
        rewriter = InsertModuleCallAfter(mul_factor_name, act_node)
        rewriter.apply(self.model)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABCMeta
from abc import abstractmethod
from typing import Optional, Type, Union

from torch.nn import Module

from brevitas.inject import ExtendedInjector
from brevitas.inject import Injector
from brevitas.proxy.runtime_quant import ActQuantProxyProtocol
from brevitas.quant import NoneActQuant

from .base import QuantProxyMixin

ActQuantType = Union[ActQuantProxyProtocol, Type[Injector], Type[ExtendedInjector]]


class QuantInputMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(self, input_quant: Optional[ActQuantType], **kwargs):
        QuantProxyMixin.__init__(
            self,
            quant=input_quant,
            proxy_protocol=ActQuantProxyProtocol,
            none_quant_injector=NoneActQuant,
            proxy_prefix='input_',
            kwargs_prefix='input_',
            input_act_impl=None,
            input_passthrough_act=True,
            **kwargs)


class QuantOutputMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(self, output_quant: Optional[ActQuantType], **kwargs):
        QuantProxyMixin.__init__(
            self,
            quant=output_quant,
            proxy_protocol=ActQuantProxyProtocol,
            none_quant_injector=NoneActQuant,
            proxy_prefix='output_',
            kwargs_prefix='output_',
            output_act_impl=None,
            output_passthrough_act=True,
            **kwargs)


class QuantNonLinearActMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(
            self,
            act_impl: Optional[Type[Module]],
            passthrough_act: bool,
            act_quant: Optional[ActQuantType],
            act_proxy_prefix='act_',
            act_kwargs_prefix='',
            **kwargs):
        prefixed_kwargs = {
            act_kwargs_prefix + 'act_impl': act_impl,
            act_kwargs_prefix + 'passthrough_act': passthrough_act}
        QuantProxyMixin.__init__(
            self,
            quant=act_quant,
            proxy_prefix=act_proxy_prefix,
            kwargs_prefix=act_kwargs_prefix,
            proxy_protocol=ActQuantProxyProtocol,
            none_quant_injector=NoneActQuant,
            **prefixed_kwargs,
            **kwargs)



import torch
import torch.nn as nn

import brevitas.nn as qnn
from brevitas.quant import Int8WeightPerTensorFloat, Int8WeightPerTensorFloatMSE
from brevitas.core.scaling import ScalingImplType

torch.manual_seed(0)

class WQMSE(Int8WeightPerTensorFloatMSE):
    bit_width=8
    #scaling_impl_type=ScalingImplType.PARAMETER_FROM_STATS

class WQ(Int8WeightPerTensorFloat):
    bit_width=8
    #scaling_impl_type=ScalingImplType.PARAMETER_FROM_STATS

with torch.no_grad():
    in_features = 16
    out_features = 32
    batch_size = 100

    x = 2*torch.rand((batch_size,in_features)) - 1.
    l = nn.Linear(in_features, out_features)
    l.eval()
    y = l(x)

    #ql = qnn.QuantLinear(in_features, out_features, weight_quant=WQ) # l2_err = 0.0054
    ql = qnn.QuantLinear(in_features, out_features, weight_quant=WQMSE) # l2_err = 298.4549
    ql.train()
    ql(x)
    ql.load_state_dict(l.state_dict(), strict=False)
    ql.eval()
    qy = ql(x)
    print(ql.quant_weight().scale)

    l2_err = ((y - qy)**2).sum()
    print(l2_err)


"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
"""

import torch
from tqdm import tqdm

from brevitas.graph.calibrate import calibration_mode


@torch.no_grad()
def apply_calibration(model, dataloader):
    with calibration_mode(model):
        for inps in tqdm(dataloader):
            model(**inps)



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
"""

from abc import ABC
from abc import abstractmethod
from contextlib import contextmanager
import warnings

import numpy as np
import torch

from brevitas.export.common.handler.base import BaseHandler
from brevitas.export.manager import _set_layer_export_handler
from brevitas.export.manager import _set_layer_export_mode
from brevitas.export.manager import _set_proxy_export_handler
from brevitas.export.manager import _set_proxy_export_mode
from brevitas.export.manager import BaseManager
from brevitas.function.ops import max_int
from brevitas.function.ops import min_int
from brevitas.nn import QuantLinear
from brevitas.proxy.parameter_quant import WeightQuantProxyFromInjector


# TODO: Improve Groupwise export
def clip_kwargs(narrow, signed, bit_width):
    if narrow or bit_width != 8. and bit_width != 32.:
        if signed and (bit_width < 8. or narrow and bit_width <= 8.):
            dtype = torch.int8
        elif not signed and (bit_width < 8. or narrow and bit_width <= 8.):
            dtype = torch.uint8
        elif signed and (bit_width < 32. or narrow and bit_width <= 32.):
            dtype = torch.int32
        else:
            raise RuntimeError(f"Sign {signed} and bit width {bit_width} not supported for export.")
        return {
            'min_val': min_int(signed, narrow, bit_width).to(dtype),
            'max_val': max_int(signed, narrow, bit_width).to(dtype)}
    else:
        return None


class WeightBlockQuantHandlerBase(BaseHandler, ABC):
    handled_layer = WeightQuantProxyFromInjector

    def __init__(self):
        super(WeightBlockQuantHandlerBase, self).__init__()
        self.int_weight = None
        self.scale = None
        self.zero_point = None
        self.bit_width = None
        self.dtype = None

    def scaling_impl(self, proxy_module):
        return proxy_module.tensor_quant.scaling_impl

    def zero_point_impl(self, proxy_module):
        return proxy_module.tensor_quant.zero_point_impl

    def bit_width_impl(self, proxy_module):
        return proxy_module.tensor_quant.msb_clamp_bit_width_impl

    def export_scale(self, proxy_module, bit_width):
        scaling_impl = self.scaling_impl(proxy_module)
        int_scaling_impl = proxy_module.tensor_quant.int_scaling_impl
        int_threshold = int_scaling_impl(bit_width)
        threshold = scaling_impl.wrapped_scaling_impl.stats_scaling_impl(
            scaling_impl.wrapped_scaling_impl.parameter_list_stats())
        return threshold / int_threshold

    def export_zero_point(self, proxy_module, scale, bit_width):
        zero_point_impl = self.zero_point_impl(proxy_module)
        return zero_point_impl.unexpanded_zero_point(scale, bit_width)

    @abstractmethod
    def prepare_for_export(self, module):
        pass

    @abstractmethod
    def forward(self, x):
        pass


class WeightBlockQuantProxyHandler(WeightBlockQuantHandlerBase):

    def __init__(self):
        super().__init__()
        self.expanded_groupwise_shape = None
        self.reshaped_groupwise_shape = None
        self.expanded_zero_point_shape = None
        self.reshaped_zero_point_shape = None

    def prepare_for_export(self, module):
        assert len(module.tracked_module_list) == 1, "Shared quantizers not supported."
        self.bit_width = self.bit_width_impl(module)()
        assert self.bit_width <= 8., "Only 8b or lower is supported."
        quant_layer = module.tracked_module_list[0]
        quant_weight = quant_layer.quant_weight()
        signed = module.is_signed
        self.int_dtype = torch.int8 if signed else torch.uint8
        self.dtype = quant_weight.value.dtype
        self.scale = self.export_scale(module, self.bit_width).detach()
        self.expanded_groupwise_shape = self.scaling_impl(module).expanded_groupwise_shape
        self.reshaped_groupwise_shape = self.scaling_impl(module).reshaped_groupwise_shape
        if (quant_weight.zero_point != 0.).any():
            self.zero_point = self.export_zero_point(module, self.scale, self.bit_width).detach()
            self.expanded_zero_point_shape = self.zero_point_impl(module).expanded_zero_point_shape
            self.reshaped_zero_point_shape = self.zero_point_impl(module).reshaped_zero_point_shape
        else:
            self.zero_point = None

        self.clip_kwargs = clip_kwargs(
            module.is_narrow_range, module.is_signed, quant_weight.bit_width)

    def forward(self, x):
        scale = self.scale
        zero_point = self.zero_point
        bit_width = self.bit_width
        # If zero point is not defined, it's all zeros
        if self.zero_point is None:
            zero_point = torch.zeros_like(scale)
        else:
            zero_point = self.zero_point

        # QCDQ
        x = x.view(self.expanded_groupwise_shape)
        x = torch.round((x / scale) + zero_point).type(self.int_dtype)
        if self.clip_kwargs is not None:
            x = torch.clip(x, min=self.clip_kwargs['min_val'], max=self.clip_kwargs['max_val'])
        x = (x.type(self.dtype) - zero_point) * scale

        # Fix shape post quantization
        scale = scale.expand(self.expanded_groupwise_shape).contiguous().view(
            self.reshaped_groupwise_shape)
        # If zero_point is not defined, propagate same shape as scale
        if self.zero_point is None:
            zero_point = torch.zeros_like(scale).type(self.int_dtype)
        else:
            zero_point = zero_point.expand(self.expanded_zero_point_shape).contiguous().view(
                self.reshaped_zero_point_shape).type(self.int_dtype)
        x = x.view(self.reshaped_groupwise_shape)

        return x, scale, zero_point, bit_width


class LinearWeightBlockQuantHandler(WeightBlockQuantHandlerBase, ABC):
    handled_layer = QuantLinear

    def __init__(self):
        super(LinearWeightBlockQuantHandler, self).__init__()
        self.group_size = None

    def pack_int_weights(self, bit_width, int_weights):
        assert int_weights.dtype in [torch.int8, torch.uint8], "Packing requires (u)int8 input."
        if bit_width == 8:
            return int_weights
        elif bit_width == 4 or bit_width == 2:
            packed_int_weights = torch.zeros(
                (int_weights.shape[0], int_weights.shape[1] * bit_width // 8),
                device=int_weights.device,
                dtype=int_weights.dtype)
            i = 0
            for column in range(packed_int_weights.shape[1]):
                # Compared to the reference below we don't transpose the matrix and we pack into 8b data rather than 32b
                # https://github.com/qwopqwop200/GPTQ-for-LLaMa/blob/05781593c818d4dc8adc2d32c975e83d17d2b9a8/quant/quant_linear.py#L346
                for j in range(i, i + (8 // bit_width)):
                    shift_factor = (bit_width * (j - i))
                    packed_int_weights[:, column] |= int_weights[:, j] << shift_factor
                i += 8 // bit_width
            return packed_int_weights

        # pack 3b values into 3 bytes, 5b values into 5 bytes, 6b values into 4 bytes
        elif bit_width == 3 or bit_width == 5 or bit_width == 6:
            padding = (int_weights.shape[1] * bit_width) % 8
            if padding > 0:
                warnings.warn(
                    f"Weight tensor does not divide by {bit_width}, zero-padding columns by {padding}."
                )
            packed_int_weights = torch.zeros(
                (int_weights.shape[0], (int_weights.shape[1] * bit_width + padding) // 8),
                device=int_weights.device,
                dtype=int_weights.dtype)

            def lcm(x, y):
                from fractions import gcd
                return x * y // gcd(x, y)

            num_packed_bits = lcm(bit_width, 8)
            num_packed_bytes = num_packed_bits // 8
            num_packed_elems = num_packed_bits // bit_width

            i = 0
            for column in range(0, packed_int_weights.shape[1], num_packed_bytes):
                # cast to uint8 since it's the only dtype supported by unpackbits
                # the bit-wise representation of int8 values isn't affected
                bits_to_unpack = int_weights[:, i:i + num_packed_elems].numpy().astype(np.uint8)
                unpacked_bits = np.unpackbits(bits_to_unpack, axis=1)
                unpacked_bits = unpacked_bits.reshape(unpacked_bits.shape[0], -1, 8)
                unpacked_bits = unpacked_bits[:, :, -bit_width:]
                unpacked_bits = unpacked_bits.reshape(unpacked_bits.shape[0], -1)
                packed_bits = np.packbits(unpacked_bits, axis=1)
                packed_int_weights[:, column:column +
                                   num_packed_bytes] |= torch.from_numpy(packed_bits)
                i += num_packed_elems
            return packed_int_weights
        else:
            raise ValueError(f"Bit width {bit_width} not supported.")

    def prepare_for_export(self, module):
        self.bit_width = self.bit_width_impl(module.weight_quant)()
        assert self.bit_width <= 8., "Only 8b or lower is supported."
        quant_weight = module.quant_weight()
        self.bias = module.bias
        self.scale = self.export_scale(module.weight_quant, self.bit_width)
        if (quant_weight.zero_point != 0.).any():
            self.zero_point = self.export_zero_point(
                module.weight_quant, self.scale, self.bit_width)
        else:
            # if there is no zero-point, export zeroes in the shape of scale
            self.zero_point = torch.zeros_like(self.scale)
        self.group_size = module.weight_quant.quant_injector.block_size
        self.bit_width = int(self.bit_width.cpu().item())
        self.int_weight = self.pack_int_weights(self.bit_width, quant_weight.int().detach())

    @abstractmethod
    def forward(self, x):
        pass


class BlockQuantProxyLevelManager(BaseManager):

    handlers = [WeightBlockQuantProxyHandler]

    @classmethod
    def set_export_handler(cls, module):
        _set_proxy_export_handler(cls, module)


def block_quant_layer_level_manager(export_handlers):

    class BlockQuantLayerLevelManager(BaseManager):
        handlers = export_handlers

        @classmethod
        def set_export_handler(cls, module):
            _set_layer_export_handler(cls, module)

    return BlockQuantLayerLevelManager


@contextmanager
def brevitas_proxy_export_mode(model, export_manager):
    is_training = model.training
    model.eval()
    model.apply(export_manager.set_export_handler)
    _set_proxy_export_mode(model, enabled=True)
    try:
        yield model
    finally:
        _set_proxy_export_mode(model, enabled=False)
        model.train(is_training)


@contextmanager
def brevitas_layer_export_mode(model, export_manager):
    is_training = model.training
    model.eval()
    model.apply(export_manager.set_export_handler)
    _set_layer_export_mode(model, enabled=True)
    try:
        yield model
    finally:
        _set_layer_export_mode(model, enabled=False)
        model.train(is_training)


def replace_call_fn_target(graph_model, src, target):
    for node in graph_model.graph.nodes:
        if node.op == "call_function" and node.target is src:
            node.target = target
    graph_model.graph.lint()
    graph_model.recompile()



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
"""

import argparse
import sys
from warnings import warn

import numpy as np
from optimum.amd.brevitas.accelerate_utils import offload_model
from optimum.amd.brevitas.accelerate_utils import remove_hooks
from optimum.amd.brevitas.data_utils import compute_perplexity
from optimum.exporters.onnx import onnx_export_from_model
import torch
from transformers import AutoModelForCausalLM
from transformers import AutoTokenizer

from brevitas.export import export_torch_qcdq
from brevitas.export.onnx.standard.qcdq.manager import StdQCDQONNXManager
from brevitas.graph.quantize import layerwise_quantize
from brevitas_examples.common.generative.quantize import generate_quant_maps
from brevitas_examples.common.generative.quantize import generate_quantizers
from brevitas_examples.common.parse_utils import quant_format_validator
from brevitas_examples.llm.llm_quant.bias_corr import apply_bias_correction
from brevitas_examples.llm.llm_quant.calibrate import apply_calibration
from brevitas_examples.llm.llm_quant.data_utils import get_dataset_for_model
from brevitas_examples.llm.llm_quant.equalize import apply_act_equalization
from brevitas_examples.llm.llm_quant.equalize import apply_weight_equalization
from brevitas_examples.llm.llm_quant.export import BlockQuantProxyLevelManager
from brevitas_examples.llm.llm_quant.export import brevitas_proxy_export_mode
from brevitas_examples.llm.llm_quant.gpxq import apply_gpfq
from brevitas_examples.llm.llm_quant.gpxq import apply_gptq
from brevitas_examples.llm.llm_quant.ln_affine_merge import apply_layernorm_affine_merge
from brevitas_examples.llm.llm_quant.prepare_for_quantize import add_zero_bias_to_linear
from brevitas_examples.llm.llm_quant.prepare_for_quantize import replace_mha_with_quantizable_layers
from brevitas_examples.llm.llm_quant.run_utils import CastFloat16ToFloat32
from brevitas_examples.llm.llm_quant.run_utils import get_fx


def set_seed(seed):
    np.random.seed(seed)
    torch.random.manual_seed(seed)


def model_export(model, ref_input, args):
    if args.export_target == 'sharded_torchmlir_group_weight':
        from brevitas_examples.llm.llm_quant.sharded_mlir_group_export import \
            sharded_weight_group_export
        sharded_weight_group_export(model, no_custom_packed_export=True)
    elif args.export_target == 'sharded_packed_torchmlir_group_weight':
        from brevitas_examples.llm.llm_quant.sharded_mlir_group_export import \
            sharded_weight_group_export
        sharded_weight_group_export(model, no_custom_packed_export=False)
    elif args.export_target == 'onnx_qcdq':
        if args.weight_quant_granularity == 'per_group':
            export_manager = BlockQuantProxyLevelManager
        else:
            export_manager = StdQCDQONNXManager
            export_manager.change_weight_export(export_weight_q_node=True)

        print(f"Exporting the model in ./{args.export_prefix}")
        with torch.no_grad(), brevitas_proxy_export_mode(model, export_manager=export_manager):
            onnx_export_from_model(
                model,
                f"./{args.export_prefix}",
                task="text-generation-with-past",
                do_validation=False)
    elif args.export_target == 'torch_qcdq':
        export_torch_qcdq(model, ref_input['input_ids'], export_path=f"{args.export_prefix}.pt")


def validate(args):
    if not args.no_quantize:
        if args.gptq and args.gpfq:
            warn("Both GPTQ and GPFQ are enabled.")
        if args.export_target is not None:
            assert args.input_quant_format == 'int', "Only integer quantization supported for export currently."
        if args.export_target is not None and args.input_bit_width is not None:
            assert args.input_scale_type == 'static', "Only static scale supported for export currently."
        if args.export_target == 'sharded_torchmlir_group_weight':
            assert args.weight_quant_granularity == 'per_group', "Sharded torch group export requires per group weight quant."
            assert args.input_bit_width is None, "Sharded torch group weight export doesn't support input quant."
            assert not args.quantize_weight_zero_point, "Quantized weight zero point not supported."
        if args.export_target == 'sharded_packed_torchmlir_group_weight':
            assert args.weight_quant_granularity == 'per_group', "Sharded torch group export requires per group weight quant."
            assert args.input_bit_width is None, "Sharded packed torch group weight export doesn't support input quant."
            assert not args.quantize_weight_zero_point, "Quantized weight zero point not supported."
        if args.export_target == 'onnx_qcdq':
            if args.weight_quant_granularity == 'per_group':
                assert args.input_bit_width is None, "ONNX QCDQ per_group quantization requires no input quantization"
            if args.weight_quant_type == 'asym':
                assert args.quantize_weight_zero_point, "Quantized weight zero point required."
            if args.input_bit_width is not None and args.input_quant_type == 'asym':
                assert args.quantize_input_zero_point, "Quantized input zero point required."
        if args.export_target == 'torch_qcdq':
            assert args.weight_quant_granularity != 'per_group', "TorchScript QCDQ export doesn't support group weight quantization."
            if args.weight_quant_type == 'asym':
                assert args.quantize_weight_zero_point, "Quantized weight zero point required."
            if args.input_bit_width is not None and args.input_quant_type == 'asym':
                assert args.quantize_input_zero_point, "Quantized input zero point required."
        if args.input_bit_width and args.input_scale_type == 'static':
            assert args.act_calibration, "Static input quantization is being applied without activation calibration. Set --act-calibration."
        if (args.weight_equalization or args.act_equalization == 'fx'):
            if args.replace_mha:
                assert args.export_target != 'onnx_qcdq', "Cannot export ONNX QCDQ with FX + MHA replacing"
            else:
                assert args.export_target != 'torch_qcdq', "Cannot export Torch QCDQ with FX"

    if not args.fuse_sequences:
        # 350 is approximately the 99% percentile for the sequence length in WikiText2 (train partition, using AutoTokenizer)
        if args.seqlen >= 350:
            warn(
                "Data loading can take a long time or, potentially, enter an infinite loop. Consider setting --args.fuse_sequences "
                "or decreasing the sequence length (seqlen)")


def main(args):
    validate(args)
    set_seed(args.seed)

    if args.export_prefix is None:
        args.export_prefix = f"{args.model.replace('/', '--')}"

    if args.no_float16:
        dtype = torch.float32
    else:
        dtype = torch.float16

    kwargs = {"torch_dtype": dtype}

    if args.export_target == 'torch_qcdq':
        kwargs['torchscript'] = True

    print("Model loading...")
    model = AutoModelForCausalLM.from_pretrained(args.model, **kwargs)
    print("Model loaded.")
    model.eval()
    tokenizer = AutoTokenizer.from_pretrained(args.model)
    float_ppl = None
    quant_ppl = None

    if args.load_awq:
        from brevitas_examples.llm.llm_quant.awq.pre_quant import apply_awq
        awq_results = torch.load(args.load_awq, map_location="cpu")
        with CastFloat16ToFloat32():
            apply_awq(model, awq_results)

    require_fx = True if args.weight_equalization or args.act_equalization == 'fx' or args.ln_affine_merge else False

    # Load the data for calibration and evaluation.
    calibration_loader = get_dataset_for_model(
        args.model,
        dataset_name=args.dataset,
        tokenizer=tokenizer,
        nsamples=args.nsamples,
        seqlen=args.seqlen,
        split="train",
        seed=args.seed,
        require_fx=require_fx,
        device=None,
        fuse_sequences=args.fuse_sequences,
    )

    validation_loader = get_dataset_for_model(
        args.model,
        dataset_name=args.dataset,
        tokenizer=tokenizer,
        nsamples=args.nsamples,
        seqlen=args.seqlen,
        split="validation",
        seed=args.seed,
        require_fx=require_fx,
        device=None,
        fuse_sequences=args.fuse_sequences,
    )

    device = next(iter(model.parameters())).device
    print("Data loaded.")

    if args.eval:
        assert args.export_target != 'torch_qcdq', "TorchScript QCDQ export and Evaluation simultaneously"
        print("Float model eval...")
        model = offload_model(model)
        float_ppl = compute_perplexity(
            model, validation_loader, context_length=args.seqlen // 2, tokenizer=tokenizer)
        remove_hooks(model)
        print(f"Float perplexity ({args.dataset}): {float_ppl:.3f}")

    if require_fx:
        model = get_fx(model)

    # Apply LN affine merging before inserting MHA layers
    # since currently there is support only for merging into Linear
    if args.ln_affine_merge:
        print("Apply LN affine merge...")
        apply_layernorm_affine_merge(model, dtype)
        print("LN affine merge applied.")

    # Insert standard MHA layers when performing fx based weight/act equalization to avoid dealing
    # with all the variability in HF implementations
    if args.replace_mha:
        print("Replace HF MHA with quantizable variants...")
        model = replace_mha_with_quantizable_layers(model, dtype)
        print("Replacing done.")

    if args.weight_equalization:
        print("Apply weight equalization...")
        # In case of float16 model, we need to offload to account for missing ops
        model = offload_model(model)
        apply_weight_equalization(model)
        remove_hooks(model)
        print("Weight equalization applied.")

    if args.act_equalization is not None:
        offload_model(model)
        print("Apply act equalization (SmoothQuant)...")
        apply_act_equalization(model, args.act_equalization, calibration_loader)
        print("Act equalization applied.")
        remove_hooks(model)

    if not args.no_quantize:
        name_blacklist = []
        print("Applying model quantization...")
        linear_input_quant, weight_quant, input_quant, q_scaled_quant, k_transposed_quant, v_quant, attn_output_weights_quant = generate_quantizers(
            dtype=dtype,
            weight_bit_width=args.weight_bit_width,
            weight_param_method=args.weight_param_method,
            weight_scale_precision=args.weight_scale_precision,
            weight_quant_type=args.weight_quant_type,
            weight_quant_granularity=args.weight_quant_granularity,
            weight_group_size=args.weight_group_size,
            weight_group_dim=args.weight_group_dim,
            quantize_weight_zero_point=args.quantize_weight_zero_point,
            weight_quant_format=args.weight_quant_format,
            input_bit_width=args.input_bit_width,
            input_quant_format=args.input_quant_format,
            input_scale_precision=args.input_scale_precision,
            input_scale_type=args.input_scale_type,
            input_param_method=args.input_param_method,
            input_quant_type=args.input_quant_type,
            input_quant_granularity=args.input_quant_granularity,
            input_group_size=args.input_group_size,
            quantize_input_zero_point=args.quantize_input_zero_point,
            device=device)
        layer_map = generate_quant_maps(
            linear_input_quant=linear_input_quant,
            weight_quant=weight_quant,
            input_quant=input_quant,
            q_scaled_quant=q_scaled_quant,
            k_transposed_quant=k_transposed_quant,
            v_quant=v_quant,
            attn_output_weights_quant=attn_output_weights_quant,
            dtype=dtype,
            device=device,
            input_quant_format=args.input_quant_format,
            quantize_embedding=False)
        if not args.quantize_last_layer:
            name_blacklist += ["lm_head", "embed_out"]
        model = layerwise_quantize(
            model=model, compute_layer_map=layer_map, name_blacklist=name_blacklist)
        # Tie back first/last layer weights in case they got untied
        print("Model quantization applied.")

    # If any equalization has taken places, the embedding layer and the fully connected one are
    # not tied anymore, and they need to be treated as standalone, separate layers.
    # In all other cases we can tie them back so to preserve memory.
    if args.act_equalization is None and not require_fx:
        model.tie_weights()

    if args.bias_corr:
        model = add_zero_bias_to_linear(model)

    model = offload_model(model)

    if args.act_calibration:
        print("Apply act calibration...")
        apply_calibration(model, calibration_loader)
        print("Act calibration applied.")

    if args.gptq:
        print("Applying GPTQ...")
        apply_gptq(
            model,
            calibration_loader,
            act_order=args.gpxq_act_order,
            use_quant_activations=args.gpxq_use_quant_activations,
            create_weight_orig=args.gpxq_create_weight_orig)
        print("GPTQ applied.")

    if args.gpfq:
        print("Applying GPFQ...")
        apply_gpfq(model, calibration_loader, act_order=args.gpxq_act_order)
        print("GPFQ applied.")

    if args.bias_corr:
        print("Applying bias correction...")
        apply_bias_correction(model, calibration_loader)
        print("Bias correction applied.")

    if args.eval:
        print("Model eval...")
        quant_ppl = compute_perplexity(
            model, validation_loader, context_length=args.seqlen // 2, tokenizer=tokenizer)
        print(f"Quantized perplexity ({args.dataset}): {quant_ppl:.3f}")
    remove_hooks(model)

    if args.checkpoint_name is not None:
        print(f"Saving checkpoint to {args.checkpoint_name}")
        torch.save(model.state_dict(), args.checkpoint_name)

    if args.export_target:
        print(f"Export to {args.export_target}")
        # Currently we always export on CPU with a float32 container to avoid float16 CPU errors
        model = model.to(dtype=torch.float32)
        model_export(model, calibration_loader[0], args)

    return float_ppl, quant_ppl, model


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--model',
        type=str,
        default="facebook/opt-125m",
        help='HF model name. Default: facebook/opt-125m.')
    parser.add_argument(
        '--seed', type=int, default=0, help='Seed for sampling the calibration data. Default: 0.')
    parser.add_argument(
        '--nsamples',
        type=int,
        default=128,
        help='Number of calibration data samples. Default: 128.')
    parser.add_argument('--seqlen', type=int, default=2048, help='Sequence length. Default: 2048.')
    parser.add_argument('--eval', action='store_true', help='Eval model PPL on the chosen Dataset.')
    parser.add_argument(
        '--dataset',
        type=str,
        choices=['wikitext2', 'c4'],
        default='wikitext2',
        help='Dataset to use for quantization (default: %(default)s)')
    parser.add_argument(
        '--weight-bit-width', type=int, default=8, help='Weight bit width. Default: 8.')
    parser.add_argument(
        '--weight-param-method',
        type=str,
        default='stats',
        choices=['stats', 'mse', 'hqo'],
        help='How scales/zero-point are determined. Default: stats.')
    parser.add_argument(
        '--weight-scale-precision',
        type=str,
        default='float_scale',
        choices=['float_scale', 'po2_scale'],
        help='Whether scale is a float value or a po2. Default: po2.')
    parser.add_argument(
        '--weight-quant-type',
        type=str,
        default='sym',
        choices=['sym', 'asym'],
        help='Weight quantization type. Default: asym.')
    parser.add_argument(
        '--weight-quant-format',
        type=quant_format_validator,
        default='int',
        help=
        'Weight quantization type. Either int or eXmY, with X+Y==weight_bit_width-1. It\'s possible to add float_ocp_ or float_fnuz_ before the exponent/mantissa bitwidth. Default: int.'
    )
    parser.add_argument(
        '--weight-quant-granularity',
        type=str,
        default='per_group',
        choices=['per_channel', 'per_tensor', 'per_group'],
        help='Granularity for scales/zero-point of weights. Default: per_group.')
    parser.add_argument(
        '--weight-group-dim',
        type=int,
        default=None,
        choices=[1, 0],
        help='Override default group_dim for groupsize quantization. Default: layer-dependant')
    parser.add_argument(
        '--weight-group-size',
        type=int,
        default=128,
        help='Group size for per_group weight quantization. Default: 128.')
    parser.add_argument(
        '--quantize-weight-zero-point', action='store_true', help='Quantize weight zero-point.')
    parser.add_argument(
        '--input-bit-width',
        type=int,
        default=None,
        help='Input bit width. Default: None (disables input quantization).')
    parser.add_argument(
        '--input-quant-format',
        type=quant_format_validator,
        default='int',
        help=
        'Input quantization type. Either int or eXmY, with X+Y==weight_bit_width-1. It\'s possible to add float_ocp_ or float_fnuz_ before the exponent/mantissa bitwidth. Default: int.'
    )
    parser.add_argument(
        '--input-param-method',
        type=str,
        default='stats',
        choices=['stats', 'mse'],
        help=
        'How scales/zero-point are determined. Default: stats (percentile for static, absmax or minmax for dynamic).'
    )
    parser.add_argument(
        '--input-scale-precision',
        type=str,
        default='float_scale',
        choices=['float_scale', 'po2_scale'],
        help='Whether input scale is a float value or a po2. Default: float.')
    parser.add_argument(
        '--input-scale-type',
        type=str,
        default='static',
        choices=['static', 'dynamic', 'no_scale'],
        help='Whether input scale is a static value or a dynamic value.')
    parser.add_argument(
        '--input-quant-type',
        type=str,
        default='asym',
        choices=['sym', 'asym'],
        help='Input quantization type. Default: asym.')
    parser.add_argument(
        '--input-quant-granularity',
        type=str,
        default='per_tensor',
        choices=['per_tensor', 'per_row', 'per_group'],
        help='Granularity for scales/zero-point of inputs. Default: per_tensor.')
    parser.add_argument(
        '--input-group-size',
        type=int,
        default=64,
        help='Group size for per_group input quantization. Default: 64.')
    parser.add_argument(
        '--quantize-input-zero-point', action='store_true', help='Quantize input zero-point.')
    parser.add_argument(
        '--quantize-last-layer', action='store_true', help='Quantize last nn.Linear layer.')
    parser.add_argument('--gptq', action='store_true', help='Apply GPTQ.')
    parser.add_argument('--gpfq', action='store_true', help='Apply GPFQ.')
    parser.add_argument(
        '--gpxq-act-order', action='store_true', help='Apply GPXQ activation ordering.')
    parser.add_argument(
        '--gpxq-use-quant-activations',
        action='store_true',
        help='Use quantized activations in GPXQ.')
    parser.add_argument(
        '--gpxq-create-weight-orig', action='store_true', help='Create weight_orig in GPXQ.')
    parser.add_argument(
        '--act-calibration', action='store_true', help='Apply activation calibration.')
    parser.add_argument('--bias-corr', action='store_true', help='Apply bias correction.')
    parser.add_argument('--ln-affine-merge', action='store_true', help='Merge LN affine params.')
    parser.add_argument('--no-quantize', action='store_true', help='Disable quantization.')
    parser.add_argument(
        '--no-float16',
        action='store_true',
        help='Disable float16 as base datatype and switch to float32.')
    parser.add_argument(
        '--replace-mha',
        action='store_true',
        help='Replace HuggingFace Attention with a quantizable version')
    parser.add_argument(
        '--weight-equalization',
        action='store_true',
        help='Apply weight equalization. Relevant to ReLU based models (e.g. OPT).')
    parser.add_argument(
        '--act-equalization',
        default=None,
        choices=[None, 'layerwise', 'fx'],
        help='Apply activation equalization (SmoothQuant). Layerwise introduces standalone mul nodes,'
        'while fx merges them whenever possible into previous tensors, which is possible on ReLU based models (e.g. OPT).'
    )
    parser.add_argument('--load-awq', type=str, default=None, help="Load the awq search results.")
    parser.add_argument(
        '--export-target',
        default=None,
        choices=[
            None,
            'onnx_qcdq',
            'torch_qcdq',
            'sharded_torchmlir_group_weight',
            'sharded_packed_torchmlir_group_weight'],
        help='Model export.')
    parser.add_argument(
        '--export-prefix',
        type=str,
        default=None,
        help=
        "Path prefix to use for the various export flows. If None, a path will be derived from the model name (default: %(default)s)"
    )
    parser.add_argument(
        '--checkpoint-name',
        type=str,
        default=None,
        help="Filename to save checkpoint. If `None`, no checkpoint is saved (default: %(default)s)"
    )
    parser.add_argument(
        "--fuse-sequences",
        action="store_true",
        default=False,
        help=
        "Whether to merge the dataset sequences in case they are shorter than the requested number of samples per sequence. This is useful in case you would like to quantize or evaluate on long sequences (default: %(default)s).",
    )
    return parser.parse_args(args)


if __name__ == '__main__':
    args = parse_args(sys.argv[1:])
    main(args)



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
SPDX-License-Identifier: MIT
"""

import torch
from torch import nn

from brevitas.graph.equalize import _is_reshaping_op
from brevitas.graph.equalize import _is_scale_invariant_module
from brevitas.graph.utils import get_module
from brevitas_examples.llm.llm_quant.run_utils import cast_to_float32


def replace_bias(next_module, new_bias):
    new_bias = new_bias.view(-1)
    if next_module.bias is not None:
        next_module.bias.data.copy_(new_bias)
    else:
        new_bias = new_bias.to(next_module.weight.device).to(next_module.weight.dtype)
        next_module.register_parameter('bias', torch.nn.Parameter(new_bias))


def _merge_ln(layer_norm, next_module, scale_bias_by_weight):
    if not layer_norm.elementwise_affine:
        return False
    if not isinstance(next_module, nn.Linear):
        return False
    view_shape = (1, -1)
    # Merge weight
    if scale_bias_by_weight:
        layer_norm.bias.data /= layer_norm.weight.data
    # We can't do an inplace update as some layers we merge into like lm_head might share the weight tensor
    scale = layer_norm.weight.data.view(view_shape).expand_as(next_module.weight)
    next_module.weight = torch.nn.Parameter(next_module.weight.clone() * scale)
    # Merge bias, new_bias includes the bias of next_module by going through its fwd
    inp = layer_norm.bias.data.view(view_shape)
    new_bias = next_module(inp)
    replace_bias(next_module, new_bias)
    return True


def merge_layernorm_affine_params(graph_model):
    merged_dict = {}
    merged_into_layers = []
    scaled_biases = set()
    for node in graph_model.graph.nodes:
        if node.op == 'call_module':
            module = get_module(graph_model, node.target)
            if isinstance(module, nn.LayerNorm):
                for next in node.users:
                    while (_is_reshaping_op(next) or _is_scale_invariant_module(graph_model, next)):
                        next = node.next
                    if next.op == 'call_module':
                        next_module = get_module(graph_model, next.target)
                        scale_bias = node.target not in scaled_biases
                        merged = _merge_ln(module, next_module, scale_bias_by_weight=scale_bias)
                        if merged:
                            print(
                                f"{module.__class__.__name__} {node.target} merged into {next.target}."
                            )
                            merged_into_layers.append(next.target)
                            scaled_biases.add(node.target)
                        if module in merged_dict:
                            merged_dict[module] &= merged
                        else:
                            merged_dict[module] = merged
                    elif next.op == 'call_method' and next.target == 'size':
                        continue
                    else:
                        raise RuntimeError(
                            f"Unsupported user node {next.op} with target {next.target}. Disable LN affine merging."
                        )
    for module, merged in merged_dict.items():
        if merged:
            # We preserve weight and bias in case they are used to merge SmoothQuant scales in fx mode later on
            module.weight.data.fill_(1.)
            module.bias.data.fill_(0.)
        else:
            raise RuntimeError(
                f"Merged only into some users: {merged_dict}. Disable LN affine merging.")
    return merged_into_layers


@torch.no_grad()
def apply_layernorm_affine_merge(graph_model, dtype):
    # We can't do fp16 tracing on CPU as many kernels are not implemented
    # So we have to cast to fp32 first, trace, apply merging, and then cast back
    with cast_to_float32(graph_model, dtype):
        merge_layernorm_affine_params(graph_model)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch

from brevitas.function.ops_ste import round_ste


def stochastic_round_ste_fn(generator):

    class StochasticRoundSteFn(torch.autograd.Function):

        @staticmethod
        def forward(ctx, x):
            floor_x = torch.floor(x)
            x_diff = torch.abs(x - floor_x)
            prob = torch.bernoulli(x_diff, generator=generator)
            out = torch.where(prob.to(torch.bool), floor_x + 1., floor_x)
            return out

        @staticmethod
        def backward(ctx, x_grad):
            return x_grad

    return StochasticRoundSteFn.apply


class StochasticRoundSte(torch.nn.Module):

    def __init__(self, deterministic_inference=True, seed=None, device=None) -> None:
        super().__init__()
        self.generator = torch.Generator(device=device)
        if seed is not None:
            self.generator.manual_seed(seed)
        self.round_fn = stochastic_round_ste_fn(self.generator)
        self.deterministic_inference = deterministic_inference
        if deterministic_inference:
            self.inference_fn = round_ste
        else:
            self.inference_fn = None

    @torch.jit.ignore
    def forward(self, x):
        if self.deterministic_inference and not self.training:
            return self.inference_fn(x)
        return self.round_fn(x)
